""" 
Classes for creating network graph models
"""
from core import BasePipeline
from utils import rando_name
from uuid import uuid4
import pandas as pd

#################################################################################################################################################################
class MinimumSpanningTreePipeline(BasePipeline):
    """ 
    Create a Pipeline for training :class:`.MinimumSpanningTree` models.

    Parameters
    ==========

    name : string
        The desired name of the Pipeline.

    corr_method : string ["pearson" | "spearman" | "kendall" ]
        Which method should we use for computing the correlation matrix?
 
       "pearson" : use the Pearson correlation coefficient

       "spearman" : use Spearman's rho

       "kendall" : use Kendall's tau

    client : :class:`slicematrixIO.connect.ConnectIO`
        Low level client for dispatching requests to SliceMatrix-IO

    Returns
    =======

    response : dict
        success or failure response to Pipeline creation request

    Examples
    ========

    Create a Pipeline for training multiple :class:`.MinimumSpanningTree` models

    >>> io = ConnectIO(api_key)
    >>> pipe = MinimumSpanningTreePipeline(client = io)
    >>> for dataframe in dataframes:
    >>>     current_model = pipe.run(dataset = dataframe, name = slicematrixIO.utils.rando_name())

    """
    def __init__(self, name, corr_method = "pearson", client = None):
        params = {"corr_method": corr_method}
        BasePipeline.__init__(self, name, "raw_mst", client, params)

    def run(self, dataset, model):
        return BasePipeline.run(self, dataset = dataset, model = model)

class MinimumSpanningTree():
    def __init__(self, dataset = None, name = None, pipeline = None, corr_method = "pearson", client = None):
        self.client  = client
        self.type     = "raw_mst"
        if dataset is not None:
            self.__full_init__(dataset, name, pipeline, corr_method, client)
        else:
            self.__lazy_init__(name)

    def __full_init__(self, dataset, name = None, pipeline = None, corr_method = "pearson", client = None):
        if name == None:
            name = rando_name()
        self.name     = name
        self.dataset  = dataset
        self.pipeline = pipeline
        self.corr_method = corr_method
        if self.pipeline == None:
            pipeline_name = rando_name()
            self.pipeline = MinimumSpanningTreePipeline(pipeline_name, corr_method, client)
        self.response = self.pipeline.run(self.dataset, self.name)
        try:
            # model will be key if success
            model = self.response['model']
            self.name = model.split("/")[-1]
        except:
            # something went wrong creating the model
            raise StandardError(self.response)

    # lazy loading for already persisted models
    def __lazy_init__(self, model_name):
        self.name     = model_name

    def nodes(self):
        """ 
        Get the names of the data points / nodes that make of the training dataset

        Returns
        =======

        nodes : list
            Data point names / indices
        """
        response = self.client.call_model(model  = self.name,
                                          type   = self.type,
                                          method = "nodes",
                                          extra_params = {})
        try:
           return response['nodes']
        except:
           raise StandardError(response)

    def rankLinks(self):
        """ 
        Rank the links by weight, if applicable

        Returns
        =======

        links : dict
            dictionary of links with associated weight, if applicable

        """
        response = self.client.call_model(model  = self.name,
                                          type   = self.type,
                                          method = "rankLinks",
                                          extra_params = {})
        try:
           return response['rankLinks']
        except:
           raise StandardError(response)

    def edges(self):
        """ 
        Get a list of all the edges in the graph model

        Returns
        =======

        edges : list
            list of all edge / link tuples. Source is edge[0] Target is edge[1]
        """
        response = self.client.call_model(model  = self.name,
                                          type   = self.type,
                                          method = "edges",
                                          extra_params = {})
        try:
           return response['edges']
        except:
           raise StandardError(response)

    def rankNodes(self, statistic = "closeness_centrality"):
        """ 
        Rank the model's nodes by the given network graph statistic / factor

        Parameters
        ==========

        statistic : string ['degree_centrality' | 'eigen_centrality' | 'closeness_centrality' | 'betweenness_centrality' | 'is_connected' |
                            'curr_flow_centrality' | 'pagerank' | 'hits' | 'communicability' | 'clustering' | 'square_clustering' |
                            'greedy_colors' | 'eccentricity' | 'clique_numbers' | 'number_of_cliques' | 'estrada_index' | 'assortivity' |
                            'transitivity' | 'avg_clustering' | 'maximal_matching' | 'max_weight_matching' | 'dispersion']

            The desired graph statistic

        Returns
        =======

        stats : array-like
            Depending on the statistic this will be an array or a single float value

        """
        extra_params = {"statistic": statistic}
        response = self.client.call_model(model  = self.name,
                                          type   = self.type,
                                          method = "rankNodes",
                                          extra_params = extra_params)
        try:
           return pd.DataFrame(response['rankNodes'], index = [statistic]).T.sort(columns = statistic)
        except:
           raise StandardError(response)

    def neighborhood(self, node):
        """ 
        Get the nearest neighbors of the given node

        Parameters
        ==========

        node : string
            The name of the target node we want to find the neighbors (shared edges)

        Returns
        =======

        neighbors : dict
            Dictionary of nearest neighbors with distances to target node

        """
        extra_params = {"node": node}
        response = self.client.call_model(model  = self.name,
                                          type   = self.type,
                                          method = "neighborhood",
                                          extra_params = extra_params)
        try:
           return response['neighborhood']
        except:
           raise StandardError(response)


#################################################################################################################################################################
class CorrelationFilteredGraphPipeline(BasePipeline):
    """ 
    Create a Pipeline for training :class:`.CorrelationFilteredGraph` models.

    CFG's are similar to MST's, in that both graph's begin with a distance matrix, but whereas
    MST's are limited to constructing a tree, CFG's draw links between a node and its closests K
    neighbors based on correlation distance. CFG's are like KNN networks, but optimized for
    using correlation distance.

    Parameters
    ==========

    name : string
        The desired name of the Pipeline.

    K : integer greater than 1, optional
        The number of nearest neighbors to use for constructing the CFG

    client : :class:`slicematrixIO.connect.ConnectIO`
        Low level client for dispatching requests to SliceMatrix-IO

    Returns
    =======

    response : dict
        success or failure response to Pipeline creation request

    Examples
    ========

    Create a Pipeline for training multiple :class:`.CorrelationFilteredGraph` models

    >>> io = ConnectIO(api_key)
    >>> pipe = CorrelationFilteredGraphPipeline(client = io)
    >>> for dataframe in dataframes:
    >>>     current_model = pipe.run(dataset = dataframe, name = slicematrixIO.utils.rando_name())
    """
    def __init__(self, name, K = 3, client = None):
        params = {"K": K}
        BasePipeline.__init__(self, name, "raw_cfg", client, params)

    def run(self, dataset, model):
        return BasePipeline.run(self, dataset = dataset, model = model)

class CorrelationFilteredGraph():
    def __init__(self, dataset = None, name = None, pipeline = None, K = 3, client = None):
        self.client  = client
        self.type     = "raw_cfg"
        if dataset is not None:
            self.__full_init__(dataset, name, pipeline, K, client)
        else:
            self.__lazy_init__(name)

    def __full_init__(self, dataset, name = None, pipeline = None, K = 3, client = None):
        if name == None:
            name = rando_name()
        self.name     = name
        self.dataset  = dataset
        self.pipeline = pipeline
        self.K        = K
        if self.pipeline == None:
            pipeline_name = rando_name()
            self.pipeline = CorrelationFilteredGraphPipeline(pipeline_name, K, client)
        self.response = self.pipeline.run(self.dataset, self.name)
        try:
            # model will be key if success
            model = self.response['model']
            self.name = model.split("/")[-1]
        except:
            # something went wrong creating the model
            raise StandardError(self.response)

    # lazy loading for already persisted models
    def __lazy_init__(self, model_name):
        self.name     = model_name

    def nodes(self):
        """ 
        Get the names of the data points / nodes that make of the training dataset

        Returns
        =======

        nodes : list
            Data point names / indices
        """
        response = self.client.call_model(model  = self.name,
                                          type   = self.type,
                                          method = "nodes",
                                          extra_params = {})
        try:
           return response['nodes']
        except:
           raise StandardError(response)

    def rankLinks(self):
        """ 
        Rank the links by weight, if applicable

        Returns
        =======

        links : dict
            dictionary of links with associated weight, if applicable

        """
        response = self.client.call_model(model  = self.name,
                                          type   = self.type,
                                          method = "rankLinks",
                                          extra_params = {})
        try:
           return response['rankLinks']
        except:
           raise StandardError(response)

    def edges(self):
        """ 
        Get a list of all the edges in the graph model

        Returns
        =======

        edges : list
            list of all edge / link tuples. Source is edge[0] Target is edge[1]
        """
        response = self.client.call_model(model  = self.name,
                                          type   = self.type,
                                          method = "edges",
                                          extra_params = {})
        try:
           return response['edges']
        except:
           raise StandardError(response)

    def rankNodes(self, statistic = "closeness_centrality"):
        """ 
        Rank the model's nodes by the given network graph statistic / factor

        Parameters
        ==========

        statistic : string ['degree_centrality' | 'eigen_centrality' | 'closeness_centrality' | 'betweenness_centrality' | 'is_connected' |
                            'curr_flow_centrality' | 'pagerank' | 'hits' | 'communicability' | 'clustering' | 'square_clustering' |
                            'greedy_colors' | 'eccentricity' | 'clique_numbers' | 'number_of_cliques' | 'estrada_index' | 'assortivity' |
                            'transitivity' | 'avg_clustering' | 'maximal_matching' | 'max_weight_matching' | 'dispersion']

            The desired graph statistic

        Returns
        =======

        stats : array-like
            Depending on the statistic this will be an array or a single float value

        """
        extra_params = {"statistic": statistic}
        response = self.client.call_model(model  = self.name,
                                          type   = self.type,
                                          method = "rankNodes",
                                          extra_params = extra_params)
        try:
           return pd.DataFrame(response['rankNodes'], index = [statistic]).T.sort(columns = statistic)
        except:
           raise StandardError(response)

    def neighborhood(self, node):
        """ 
        Get the nearest neighbors of the given node

        Parameters
        ==========

        node : string
            The name of the target node we want to find the neighbors (shared edges)

        Returns
        =======

        neighbors : dict
            Dictionary of nearest neighbors with distances to target node

        """
        extra_params = {"node": node}
        response = self.client.call_model(model  = self.name,
                                          type   = self.type,
                                          method = "neighborhood",
                                          extra_params = extra_params)
        try:
           return response['neighborhood']
        except:
           raise StandardError(response)

#################################################################################################################################################################
class NeighborNetworkGraphPipeline(BasePipeline):
    """ 
    Create a Pipeline for training :class:`.NeighborNetworkGraph` models.

    Parameters
    ==========

    name : string
        The desired name of the Pipeline.

    K : integer greater than 1, optional
        The number of nearest neighbors to use for constructing the CFG

    kernel : string, optional
        The distance kernel / metric to use in constructing the distance matrix. Default is euclidean.

    client : :class:`slicematrixIO.connect.ConnectIO`
        Low level client for dispatching requests to SliceMatrix-IO

    Returns
    =======

    response : dict
        success or failure response to Pipeline creation request

    Examples
    ========

    Create a Pipeline for training multiple :class:`.NeighborNetworkGraph` models

    >>> io = ConnectIO(api_key)
    >>> pipe = NeighborNetworkGraphPipeline(K = 5, client = io)
    >>> for dataframe in dataframes:
    >>>     current_model = pipe.run(dataset = dataframe, name = slicematrixIO.utils.rando_name())

    """
    def __init__(self, name, K = 3, kernel = "euclidean", client = None):
        params = {"K": K, 
                  "kernel": kernel}
        BasePipeline.__init__(self, name, "raw_knn_net", client, params)

    def run(self, dataset, model):
        return BasePipeline.run(self, dataset = dataset, model = model)

class NeighborNetworkGraph():
    def __init__(self, dataset = None, name = None, pipeline = None, K = 3, kernel = "euclidean", client = None):
        self.client  = client
        self.type     = "raw_knn_net"
        if dataset is not None:
            self.__full_init__(dataset, name, pipeline, K, kernel, client)
        else:
            self.__lazy_init__(name)

    def __full_init__(self, dataset, name = None, pipeline = None, K = 3, kernel = "euclidean", client = None):
        if name == None:
            name = rando_name()
        self.name     = name
        self.dataset  = dataset
        self.pipeline = pipeline
        self.K        = K
        if self.pipeline == None:
            pipeline_name = rando_name()
            self.pipeline = NeighborNetworkGraphPipeline(pipeline_name, K, kernel, client)
        self.response = self.pipeline.run(self.dataset, self.name)
        try:
            # model will be key if success
            model = self.response['model']
            self.name = model.split("/")[-1]
        except:
            # something went wrong creating the model
            raise StandardError(self.response)

    # lazy loading for already persisted models
    def __lazy_init__(self, model_name):
        self.name     = model_name

    def nodes(self):
        """ 
        Get the names of the data points / nodes that make of the training dataset

        Returns
        =======

        nodes : list
            Data point names / indices
        """
        response = self.client.call_model(model  = self.name,
                                          type   = self.type,
                                          method = "nodes",
                                          extra_params = {})
        try:
           return response['nodes']
        except:
           raise StandardError(response)

    def rankLinks(self):
        """ 
        Rank the links by weight, if applicable

        Returns
        =======

        links : dict
            dictionary of links with associated weight, if applicable

        """
        response = self.client.call_model(model  = self.name,
                                          type   = self.type,
                                          method = "rankLinks",
                                          extra_params = {})
        try:
           return response['rankLinks']
        except:
           raise StandardError(response)

    def edges(self):
        """ 
        Get a list of all the edges in the graph model

        Returns
        =======

        edges : list
            list of all edge / link tuples. Source is edge[0] Target is edge[1]
        """
        response = self.client.call_model(model  = self.name,
                                          type   = self.type,
                                          method = "edges",
                                          extra_params = {})
        try:
           return response['edges']
        except:
           raise StandardError(response)

    def rankNodes(self, statistic = "closeness_centrality"):
        """ 
        Rank the model's nodes by the given network graph statistic / factor

        Parameters
        ==========

        statistic : string ['degree_centrality' | 'eigen_centrality' | 'closeness_centrality' | 'betweenness_centrality' | 'is_connected' |
                            'curr_flow_centrality' | 'pagerank' | 'hits' | 'communicability' | 'clustering' | 'square_clustering' |
                            'greedy_colors' | 'eccentricity' | 'clique_numbers' | 'number_of_cliques' | 'estrada_index' | 'assortivity' |
                            'transitivity' | 'avg_clustering' | 'maximal_matching' | 'max_weight_matching' | 'dispersion']

            The desired graph statistic

        Returns
        =======

        stats : array-like
            Depending on the statistic this will be an array or a single float value

        """
        extra_params = {"statistic": statistic}
        response = self.client.call_model(model  = self.name,
                                          type   = self.type,
                                          method = "rankNodes",
                                          extra_params = extra_params)
        try:
           return pd.DataFrame(response['rankNodes'], index = [statistic]).T.sort(columns = statistic)
        except:
           raise StandardError(response)

    def neighborhood(self, node):
        """ 
        Get the nearest neighbors of the given node

        Parameters
        ==========

        node : string
            The name of the target node we want to find the neighbors (shared edges)

        Returns
        =======

        neighbors : dict
            Dictionary of nearest neighbors with distances to target node

        """
        extra_params = {"node": node}
        response = self.client.call_model(model  = self.name,
                                          type   = self.type,
                                          method = "neighborhood",
                                          extra_params = extra_params)
        try:
           return response['neighborhood']
        except:
           raise StandardError(response)





from core import BasePipeline
from utils import rando_name
from uuid import uuid4
import pandas as pd

#################################################################################################################################################################
class KalmanOLSPipeline(BasePipeline):
    """ 
    Create a Pipeline for training :class:`.KalmanOLS` models from input datasets

    Parameters
    ==========
    name : string
        The desired name of the Pipeline

    init_alpha : float, optional
        Initial value for alpha in OLS model (ignored if optimizations are enabled)

    init_beta : float, optional
        Initial value for beta in OLS model (ignored if optimizations are enabled)

    trans_cov : array-like, optional
        Transition covariance, shape = (2, 2)

    init_cov : array-like, optional
        Initial covariance, shape = (2, 2)

    optimizations : list, optional
        List of optimizations. Can include multiple optimizations. Default includes all:

        - 'transition_covariance'
        - 'observation_covariance'
        - 'initial_state_mean'
        - 'initial_state_covariance'

    client : :class:`slicematrixIO.connect.ConnectIO`
        Low level client for dispatching requests to SliceMatrix-IO

    Returns
    =======

    response : dict
        success or failure response to Pipeline creation request

    Examples
    ========

    Create a KalmanOLSPipeline for processing multiple datasets

    >>> io = ConnectIO(api_key)
    >>> pipe = KalmanOLSPipeline(client = io)
    >>> for dataframe in dataframes:
    >>>     current_model = pipe.run(dataset = dataframe, name = slicematrixIO.utils.rando_name())

    """
    def __init__(self, name, init_alpha = None, init_beta = None, trans_cov = None, obs_cov = None, init_cov = None, optimizations = [], client = None):
        params = {"init_alpha": init_alpha,
                  "init_beta": init_beta,
                  "trans_cov": trans_cov,
                  "obs_cov": obs_cov,
                  "init_cov": init_cov,
                  "optimizations": optimizations}
        BasePipeline.__init__(self, name, "kalman_ols", client, params)

    def run(self, dataset, model):
        """ 
        Run the Pipeline and create a new :class:`.KalmanOLS` model

        Parameters
        ==========

        dataset : pandas.DataFrame
            The dataset to pass into the Pipeline which will train a KernelPCA model using the parameters
            defined upon Pipeline creation. Pipelines are reusable sets of instructions to train a machine
            learning model.

        Returns
        =======

        response : dict
            success or failure response to model creation request
            
        """
        return BasePipeline.run(self, dataset = dataset, model = model)

class KalmanOLS():
    """ 
    Train / Reload a Kalman Filter model for online estimation of the parameters of Ordinary Least Squares (KalmanOLS)

    Parameters
    ==========

    dataset: pandas.DataFrame
        Input DataFrame. shape = (nrows, 2) where the first column is Y and the second is X in OLS model

    init_alpha : float, optional
        Initial value for alpha in OLS model (ignored if optimizations are enabled)

    init_beta : float, optional
        Initial value for beta in OLS model (ignored if optimizations are enabled)

    trans_cov : array-like, optional
        Transition covariance, shape = (2, 2)

    init_cov : array-like, optional
        Initial covariance, shape = (2, 2)

    optimizations : list, optional
        List of optimizations. Can include multiple optimizations. Default includes all:

        - 'transition_covariance'
        - 'observation_covariance'
        - 'initial_state_mean'
        - 'initial_state_covariance'

    name : string, optional
        The desired name of the model. If None then a random name will be generated

    client : :class:`slicematrixIO.connect.ConnectIO`
        Low level client for dispatching requests to SliceMatrix-IO

    pipeline : BasePipeline, optional
        Pipeline to use. Defaults to None. If None then a pipeline will be created for use in creating the model

    Returns
    =======

    model : :class`.KalmanOLS`
        Trained Kalman Filter model

    Examples
    ========

    Create a KalmanOLS model for a given dataset

    >>> sm = SliceMatrix(api_key)
    >>> kf = sm.KalmanOLS(dataset = dataframe)

    Get the current internal state of the model (i.e. current alpha and beta and covariance)

    >>> kf.getState()

    Update the model will new information, and get updated state

    >>> kf.update(X = 128.17, Y = 45.85)

    """

    def __init__(self, dataset = None, name = None, pipeline = None, init_alpha = None, init_beta = None, trans_cov = None, obs_cov = None, init_cov = None, optimizations = [], client = None):
        self.client  = client
        self.type     = "kalman_ols"
        if dataset is not None:
            self.__full_init__(dataset, name, pipeline, init_alpha, init_beta, trans_cov, obs_cov, init_cov, optimizations, client)
        else:
            self.__lazy_init__(name)

    def __full_init__(self, dataset, name = None, pipeline = None, init_alpha = None, init_beta = None, trans_cov = None, obs_cov = None, init_cov = None, optimizations = [], client = None):
        if name == None:
            name = rando_name()
        self.name     = name
        self.dataset  = dataset
        self.pipeline = pipeline
        self.init_alpha = init_alpha
        self.init_beta = init_beta
        self.trans_cov = trans_cov
        self.obs_cov   = obs_cov
        self.init_cov  = init_cov
        self.optimizations = optimizations
        if self.pipeline == None:
            pipeline_name = rando_name()
            self.pipeline = KalmanOLSPipeline(pipeline_name, init_alpha, init_beta, trans_cov, obs_cov, init_cov, optimizations, client)
        self.response = self.pipeline.run(self.dataset, self.name)
        try:
            # model will be key if success
            model = self.response['model']
            self.name = model.split("/")[-1]
        except:
            # something went wrong creating the model
            raise StandardError(self.response)

    # lazy loading for already persisted models
    def __lazy_init__(self, model_name):
        self.name     = model_name

    def getState(self):
        """ 
        Get the current internal state of the Kalman Filter OLS model

        Returns
        =======

        state : dict
            Dictionary with the current state of model i.e. 
            - means (Beta and Alpha, respectively)
            - covariance 

        """
        response = self.client.call_model(model  = self.name,
                                          type   = self.type,
                                          method = "getState",
                                          extra_params = {})
        try:
           return response['getState']
        except:
           raise StandardError(response)

    def getTrainingData(self):
        """ 
        Get the historical state of the model over time

        Returns
        =======

        history : dict
            Historical state of both mean and covariance of the model over time.
        """
        response = self.client.call_model(model  = self.name,
                                          type   = self.type,
                                          method = "getTrainingData",
                                          extra_params = {})
        try:
           return response['getTrainingData']
        except:
           raise StandardError(response)

    # update the model with new data and return updated state info
    def update(self, X, Y):
        """ 
        Step the model through a new learning iteration with new datapoints for input (X) and output (Y)

        This will permanently change the state of the model as it adjusts to new information. 

        In a distributed setting, updates to the same :class:`KalmanOLS` model are not guaranteed to be atomic 

        Parameters
        ==========

        X : float
            The newly observed value for the input of the OLS model (X)

        Y : float
            The newly observed value for the output of the OLS model (Y)

        Returns
        =======

        state : dict
            Dictionary with the current state of model i.e.
            - means (Beta and Alpha, respectively)
            - covariance

        """
        extra_params = {"Y": Y, "X": X}
        response = self.client.call_model(model  = self.name,
                                          type   = self.type,
                                          method = "update",
                                          extra_params = extra_params)
        try:
           return response['update']
        except:
           raise StandardError(response)



""" 
Distance / Similarity Matrix Models

Generalization of the correlation matrix for different metrics / kernels / similarity measures
"""
from core import BasePipeline
from utils import rando_name
from uuid import uuid4
import pandas as pd

class DistanceMatrixPipeline(BasePipeline):
    """ 
    Create a Pipeline to train :class:`.DistanceMatrix` models from input datasets

    Parameters
    ==========

    name : string, optional
        The desired name of the Pipeline.

    K : integer greater than 1, optional, ignored if geodesic == False
        The number of neighbors to use in building the geodesic distance matrix. Geodesic distance is constructed by computing the
        K Nearest Neighbors graph for the input dataset, then constructing all pairwise distances using the geodesic distance, i.e.
        the number of edges in a shortest path between two points on the graph.

    kernel : string, optional
        The distance kernel / metric to use in constructing the distance matrix. Default is euclidean.

    kernel_params : dict, optional
        Any extra parameters specific to the chosen kernel

    geodesic : boolean, optional
        Whether to create the geodesic distance matrix or the brute force pairwise distance matrix. Default is False

    client : :class:`slicematrixIO.connect.ConnectIO`
        Low level client for dispatching requests to SliceMatrix-IO

    Returns
    =======

    response : dict
        success or failure response to Pipeline creation request

    Examples
    ========

    Create a Pipeline for processing multiple datasets into :class:`.DistanceMatrix` models

    >>> io = ConnectIO(api_key)
    >>> matrix_pipe = DistanceMatrixPipeline(kernel = "correlation", client = io)
    >>> for dataframe in dataframes:
    >>>     current_model = matrix_pipe.run(dataset = dataframe, name = slicematrixIO.utils.rando_name())

    """
    def __init__(self, name, kernel = "euclidean", geodesic = False, K = 5, kernel_params = {}, client = None):
        params = {"k": K,
                  "kernel": kernel,
                  "kernel_params": kernel_params,
                  "geodesic": geodesic}
        BasePipeline.__init__(self, name, "dist_matrix", client, params)

    def run(self, dataset, model):
        """ 
        Run the Pipeline and create a new :class:`.DistanceMatrix` model

        Parameters
        ==========

        dataset : pandas.DataFrame
            The dataset to pass into the Pipeline which will train a :class:`.DistanceMatrix` model using the parameters
            defined upon Pipeline creation. Pipelines are reusable sets of instructions to train a machine
            learning model.
        """
        return BasePipeline.run(self, dataset = dataset, model = model)

class DistanceMatrix():
    """ 
    Train / Reload a :class:`.DistanceMatrix` model 

    Parameters
    ==========

    name : string, optional
        The desired name of the model. If None a random name will be generated

    K : integer greater than 1, optional, ignored if geodesic == False
        The number of neighbors to use in building the geodesic distance matrix. Geodesic distance is constructed by computing the
        K Nearest Neighbors graph for the input dataset, then constructing all pairwise distances using the geodesic distance, i.e.
        the number of edges in a shortest path between two points on the graph.

    kernel : string, optional
        The distance kernel / metric to use in constructing the distance matrix. Default is euclidean.

    kernel_params : dict, optional
        Any extra parameters specific to the chosen kernel

    geodesic : boolean, optional
        Whether to create the geodesic distance matrix or the brute force pairwise distance matrix. Default is False

    client : :class:`slicematrixIO.connect.ConnectIO`
        Low level client for dispatching requests to SliceMatrix-IO

    pipeline : string, optional
       An extant :class:`DistanceMatrixPipeline` to use for model creation. If None then one will be created

    Returns
    =======
     
    """
    def __init__(self, dataset = None, name = None, pipeline = None, K = 5, kernel = "euclidean", geodesic = False, kernel_params = {}, client = None):
        self.client  = client
        self.type     = "dist_matrix" 
        if dataset is not None:
            self.__full_init__(dataset, name, pipeline, K, kernel, geodesic, kernel_params, client)
        else:
            self.__lazy_init__(name)
    
    # full construction, i.e. start from zero and create it all...
    def __full_init__(self, dataset, name = None, pipeline = None, K = 5, kernel = "euclidean", geodesic = False, kernel_params = {}, client = None):
        if name == None:
            name = rando_name()
        self.name     = name
        self.dataset  = dataset
        self.pipeline = pipeline
        self.kernel   = kernel
        self.kernel_params = kernel_params
        self.K        = K
        self.geodesic = geodesic
        if self.pipeline == None:
            pipeline_name = rando_name()
            self.pipeline = DistanceMatrixPipeline(name = pipeline_name, 
                                                   K    = K, 
                                                   kernel = kernel,
                                                   geodesic = geodesic,
                                                   kernel_params = kernel_params, 
                                                   client = client)
        self.response = self.pipeline.run(self.dataset, self.name)
        try:
            # model will be key if success
            model = self.response['model']
            self.name = model.split("/")[-1]
        except:
            # something went wrong creating the model
            raise StandardError(self.response)

    # lazy loading for already persisted models
    def __lazy_init__(self, model_name):
        self.name     = model_name
        
    def rankDist(self, target, page = 0):
        """ 
        Get the closest datapoints to the given target

        Parameters
        ==========

        page : integer, optional
            The current page. Responses come in chunks of 100. To iterate through the full list increase the page number.

        Returns
        =======

        distances : pandas.DataFrame
            DataFrame with list of datapoints sorted by distance from target point
        """
        extra_params = {"target": target,
                        "page": page}
        response = self.client.call_model(model  = self.name,
                                          type   = self.type,
                                          method = "rankDist",
                                          extra_params = extra_params)
        try:
           return pd.DataFrame(response['rankDist'], index = ['distance']).T.sort(columns = "distance")
        except:
           raise StandardError(response)

    def getKeys(self):
        """ 
        Get the names of the datapoints in the model's training dataset

        Returns
        =======

        keys : list
            The names of the datapoints in the model's training dataset
        """
        response = self.client.call_model(model  = self.name,
                                          type   = self.type,
                                          method = "getKeys",
                                          extra_params = {})
        try:
           return response['getKeys']
        except:
           raise StandardError(response)

class LazyMatrixPipeline(BasePipeline):
    def __init__(self, name, kernel = "euclidean", kernel_params = {}, client = None):
        params = {"kernel": kernel,
                  "kernel_params": kernel_params}
        BasePipeline.__init__(self, name, "lazy_matrix", client, params)

    def run(self, dataset, model):
        return BasePipeline.run(self, dataset = dataset, model = model)

class LazyMatrix():
    def __init__(self, dataset = None, name = None, pipeline = None, kernel = "euclidean", kernel_params = {}, client = None):
        self.client  = client
        self.type     = "lazy_matrix"
        if dataset is not None:
            self.__full_init__(dataset, name, pipeline, kernel, kernel_params, client)
        else:
            self.__lazy_init__(name)

    # full construction, i.e. start from zero and create it all...
    def __full_init__(self, dataset, name = None, pipeline = None, kernel = "euclidean", kernel_params = {}, client = None):
        if name == None:
            name = rando_name()
        self.name     = name
        self.dataset  = dataset
        self.pipeline = pipeline
        self.kernel   = kernel
        self.kernel_params = kernel_params
        if self.pipeline == None:
            pipeline_name = rando_name()
            self.pipeline = LazyMatrixPipeline(name = pipeline_name,
                                               kernel = kernel,
                                               kernel_params = kernel_params,
                                               client = client)
        self.response = self.pipeline.run(self.dataset, self.name)
        try:
            # model will be key if success
            model = self.response['model']
            self.name = model.split("/")[-1]
        except:
            # something went wrong creating the model
            raise StandardError(self.response)

    # lazy loading for already persisted models
    def __lazy_init__(self, model_name):
        self.name     = model_name

    def rankDist(self, target, page = 0):
        """ 
        Get the closest datapoints to the given target

        Parameters
        ==========

        page : integer, optional
            The current page. Responses come in chunks of 100. To iterate through the full list increase the page number.

        Returns
        =======

        distances : pandas.DataFrame
            DataFrame with list of datapoints sorted by distance from target point
        """
        extra_params = {"target": target,
                        "page": page}
        response = self.client.call_model(model  = self.name,
                                          type   = self.type,
                                          method = "rankDist",
                                          extra_params = extra_params)
        try:
           return pd.DataFrame(response['rankDist'], index = ['distance']).T.sort(columns = "distance")
        except:
           raise StandardError(response)

    def getKeys(self):
        """ 
        Get the names of the datapoints in the model's training dataset

        Returns
        =======

        keys : list
            The names of the datapoints in the model's training dataset
        """
        response = self.client.call_model(model  = self.name,
                                          type   = self.type,
                                          method = "getKeys",
                                          extra_params = {})
        try:
           return response['getKeys']
        except:
           raise StandardError(response)




"""low level SliceMatrix-IO API client"""

API_STAGE = "beta"

import pandas as pd
import numpy as np
import requests
from requests.packages.urllib3.exceptions import InsecureRequestWarning
requests.packages.urllib3.disable_warnings(InsecureRequestWarning)
import json
import boto3
from StringIO import StringIO
import datetime as dt
import tempfile

region_api_map    = {"us-east-1":      "ud7p0wre43",
                     "us-west-1":      "4u0pr4ljei",
                     "eu-central-1":   "pozdfzsgae",
                     "ap-southeast-1": "gbq29whx4l"}

class ConnectIO():
  """ 
  Low Level Connection to SliceMatrix-IO

  Implements basic interface

  Parameters
  ----------
  api_key : string
      Valid SliceMatrix-IO API Key

  region : string ['us-east-1', 'us-west-1', 'eu-central-1', 'ap-southeast-1']
      Data center of choice. API Key must be valid for that specific data center.
      Latency will be lowest if client is closest to data center.

      'us-east-1': US East Coast Data Center

      'us-west-1': US West Coast Data Center

      'eu-central-1': Continental Europe Data Center

      'ap-southeast-1': South-East Asian Data Center

  Attributes
  ----------
  uploader : object
      Convienence class for uploading data to SliceMatrix-IO

  region : string

  Examples
  --------
  >>> from slicematrixIO.connect import ConnectIO
  >>> io = ConnectIO(api_key)
  >>> io.create_pipeline(...)
  >>> io.run_pipeline(...)
  >>> io.call_model(...)
      
  """
  def __init__(self, api_key, region = "us-east-1"):
    self.api_key  = api_key
    self.region   = region
    self.api      = region_api_map[region]
    self.uploader = Uploader(api_key, self.region, self.api)

  def put_df(self, name, dataframe):
    """ 
    Upload the DataFrame with desired name and get response (success | failure)

    Parameters
    ----------
    name : string
        The desired name of the DataFrame

    dataframe: pandas.DataFrame
        The DataFrame for uploading to the SliceMatrix-IO backend

    Returns
    -------
    response : dict
    """
    self.uploader.put_df(name, dataframe)

  def list_files(self):
    """ 
    Get a list of the files previously uploaded

    Returns
    -------
    file_list : list
    """
    return self.uploader.list_files()  

  def create_pipeline(self, name, type, params = {}):
    """ 
    Create a new Analytical Pipeline for distributed computation

    Parameters
    ----------
    name : string
        The desired name of the new Pipeline

    type : string [ 'raw_isomap' | 'raw_mst' | 'raw_lle' | 'raw_cfg' | 'raw_kde' | 'raw_knn_net' | 'raw_knn_classifier' | 'raw_knn_regressor' | 'raw_kpca' | 'raw_krr' | 'raw_rfr' | 'raw_laplacian' | 'raw_pnn' | 'matrix_mst' | 'matrix_kpca' | 'matrix_agg' | 'kalman_ols' | 'basic_a2d' | 'isolation_forest' | 'dist_matrix']
        The type of the Pipeline

    params: dict
        Any type specific parameters to the Pipeline in the key/val dictionary

    Returns
    -------
    response: dict

    Notes
    -----
    The basic structure of computation in SliceMatrix-IO starts with the Pipeline.
    
    Pipelines can be thought of as analytical assembly lines, running code which
    transforms a dataset from raw input data into a meaningful machine learning
    model. Each pipeline can be reused to process multiple datasets. Pipelines can
    also be run in parallel.  
    """
    url = 'https://' + self.api + '.execute-api.' + self.region + '.amazonaws.com/' + API_STAGE + '/pipelines/create'
    headers = {'x-api-key': self.api_key, 'Content-Type': 'application/json'}
    body    = {'name': name, 'type': type, 'params': params}
    r = requests.post(url, verify = False, headers = headers, data=json.dumps(body))
    return json.loads(r.text)
  
  def run_pipeline(self, name, model, type = None, dataset = None, matrix_name = None, matrix_type = None, X = None, Y = None, extra_params = {}, memory = "large"):
    """ 
    Run a Pipeline with the given dataset

    Parameters
    ----------
    name : string
        The name of the target Pipeline

    model : string
        The desired name of the model

    type : string [ 'raw_isomap' | 'raw_mst' | 'raw_lle' | 'raw_cfg' | 'raw_kde' |
                    'raw_knn_net' | 'raw_knn_classifier' | 'raw_knn_regressor' |
                    'raw_kpca' | 'raw_krr' | 'raw_rfr' | 'raw_laplacian' |
                    'raw_pnn' | 'matrix_mst' | 'matrix_kpca' | 'matrix_agg' |
                    'kalman_ols' | 'basic_a2d' | 'isolation_forest' | 'dist_matrix']
        The type of the Pipeline

    dataset : string 
        The name of the dataset to run through the Pipeline
    
    matrix_name : string
        The name of the matrix model to run through the Pipeline (for Matrix Models)

    matrix_type : string [ 'dist_matrix' | 'matrix_agg' ]  
        The type of matrix

    X : string
        The name of the X input (for multi-output regression models)

    Y : string
        The name of the Y input (for multi-output regression models)

    extra_params : dict
        Any extra parameters to pass as key / values to the Pipeline

    memory: string [ 'large']
        The size of the container (always set to large for beta)

    Returns
    -------
    response : dict

    Notes
    -----
    This is a very flexible function for running any Pipeline in the SliceMatrix-IO platform. 

    Most Pipelines will take a single dataset name as input (such as raw_isomap and raw_knn_classifier),
    whereas others will have more complex inputs. Matrix Models will take matrix_name and matrix_type
    parameters and regression models will require the names of input (X) and output (Y) training sets.

    """
    url = 'https://' + self.api + '.execute-api.' + self.region + '.amazonaws.com/' + API_STAGE + '/pipelines/run'
    headers = {'x-api-key': self.api_key, 'Content-Type': 'application/json'}
    body = {'name': name, 'model': model, 'memory': memory, 'type': type}
    if dataset != None:
        body['dataset']     = dataset
    else:
        body['dataset']     = ""
    if matrix_name != None and matrix_type != None:
        body['matrix_name'] = matrix_name
        body['matrix_type'] = matrix_type
    if X != None and Y != None:
        body['X'] = X
        body['Y'] = Y
    for param_key in extra_params.keys():
      body[param_key] = extra_params[param_key]
    #print body
    r = requests.post(url, verify = False, headers = headers, data=json.dumps(body))
    return json.loads(r.text)

  def call_model(self, model, type, method, extra_params = {}, memory = "large"):
    """ 
    Remotely call a method in a machine learning model

    Parameters
    ----------
    model : string
        The name of the model 

    type : string
        The type of the model

    method: string
        The name of the model method to call remotely. Acceptable inputs vary by Pipeline type. See Pipeline docs for more information

    extra_params : dict
        Any extra parameters to pass as key / values to the Pipeline

    memory: string [ 'large']
        The size of the container (always set to large for beta)

    Returns
    -------
    model_output: dict
    
    Notes
    -----
       

    """
    url = 'https://' + self.api + '.execute-api.' + self.region + '.amazonaws.com/' + API_STAGE + '/models/call'
    headers = {'x-api-key': self.api_key, 'Content-Type': 'application/json'}
    body    = {'model': model, 'type': type, 'function': method, 'extra_params': extra_params, 'memory': memory}
    r = requests.post(url, verify = False, headers = headers, data=json.dumps(body))
    #print r.text
    return json.loads(r.text)#[method]

  
class Uploader():
  """ 
  Object to handle uploads to SliceMatrix-IO backend

  Parameters
  ----------
  api_key : string
     Valid SliceMatrix-IO API Key

  region : string ['us-east-1', 'us-west-1', 'eu-central-1', 'ap-southeast-1']
      Data center of choice. API Key must be valid for that specific data center.
      Latency will be lowest if client is closest to data center.

      'us-east-1': US East Coast Data Center

      'us-west-1': US West Coast Data Center

      'eu-central-1': Continental Europe Data Center

      'ap-southeast-1': South-East Asian Data Center

  api : string 
      API ID

  Examples
  --------
  >>> uploader = Uploader(api_key)
  >>> uploader.put_df("my_dataframe", df)   

  """
  def __init__(self, api_key, region, api):
    self.api_key = api_key
    self.region  = region
    self.api     = api
	
  def get_upload_url(self, file_name):
    url = 'https://' + self.api + '.execute-api.' + self.region + '.amazonaws.com/' + API_STAGE + '/datasets/authorize'
    headers = {'x-api-key': self.api_key, 'Content-Type': 'application/json'}
    r = requests.post(url, verify = False, headers = headers, data = json.dumps({'name':file_name}))
    return json.loads(r.text)

  def put_df(self, name, df):
    """ 
    Upload the DataFrame with desired name and get response (success | failure)

    Parameters
    ----------
    name : string
        The desired name of the DataFrame

    df: pandas.DataFrame
        The DataFrame for uploading to the SliceMatrix-IO backend

    Returns
    -------
    response : dict
    """
    post = self.get_upload_url(name)
    files = {"file": df.to_csv()}
    response = requests.post(post["url"], data=post["fields"], files=files)
    return response
	
  def list_files(self): 
    """ 
    Get a list of the files previously uploaded

    Returns
    -------
    file_list : list
    """
    url = 'https://' + self.api + '.execute-api.' + self.region + '.amazonaws.com/' + API_STAGE + '/datasets'
    headers = {'x-api-key': self.api_key}
    r = requests.get(url, verify = False, headers = headers)
    return json.loads(r.text)




""" 
Regressors are machine learning models which learn a function between an input (X) and an output (Y).

In particular, SliceMatrix-IO offers a number of what are known as "multi-output" regression models.

This is a special type of regression which can have an output with a dimension greater than 1, useful for:

- Prediction
- Out of Sample Manifold Learning
- As a step within a classification workflow

"""
from core import BasePipeline
from utils import rando_name, r_squared
from uuid import uuid4
import pandas as pd
import numpy as np

# Random Forest Regressor pipeline #############################################################################################################################################3
class RFRegressorPipeline(BasePipeline):
    """ 
    Random Forest Regression.

    Create a Pipeline for training :class:`RFRegressor` models.

    Parameters
    ==========

    name : string
        The desired name of the Pipeline.

    n_trees : integer, greater than 0
        The number of trees to use in the regression forest

    client : :class:`slicematrixIO.connect.ConnectIO`
        Low level client for dispatching requests to SliceMatrix-IO

    Returns
    =======

    response : dict
        success or failure response to Pipeline creation request

    Examples
    ========

    Create a Pipeline for training multiple :class:`.RFRegressor` models

    >>> io = ConnectIO(api_key)
    >>> pipe = RFRegressorPipeline(n_trees = 100, client = io)
    >>> for dataframe in dataframes:
    >>>     current_model = pipe.run(dataset = dataframe, name = slicematrixIO.utils.rando_name())

    """
    def __init__(self, name, n_trees = 8, client = None):
        params = {"n_trees": n_trees,}
        BasePipeline.__init__(self, name, "raw_rfr", client, params)

    def run(self, X, Y, model):
        """ 
        Run the Pipeline and create a new :class:`RFRegressor` model

        Parameters
        ==========

        dataset : pandas.DataFrame
            The dataset to pass into the Pipeline which will train a :class:`RFRegressor` model using the
            parameters defined upon Pipeline creation. Pipelines are reusable sets of instructions to train
            a machine learning model.

        Returns
        =======

        response : dict
            success or failure response to model creation request
        """
        return BasePipeline.run(self, X = X, Y = Y, model = model)

# Random Forest Regressor model; if pipeline == None then create a pipeline for use with this model
class RFRegressor():
    """ 
    Train / Reload a :class:`.RFRegressor` model for multi-output regression

    A Random Forest Regressor finds a function which maps the input space (X) to the lower dimension output space (Y) using decision trees

    Parameters
    ==========

    X : pandas.DataFrame
        Input DataFrame. shape = (n_rows, input_features) where each row is a data point and the columns are numeric features

    Y : pandas.DataFrame
        Output DataFrame. shape = (n_rows, output_features) where output_features < input_features and each row is a data point
        and the columns are numeric features

    n_trees : integer greater than 1, optional
        The number of trees to use in construction of the Random Forest model. Default is 8 trees

    name : string, optional
        The desired name of the model. If None then a random name will be generated

    pipeline : string, optional
        An extant Pipeline to use for model creation. If None then one will be created

    client : :class:`slicematrixIO.connect.ConnectIO`
        Low level client for dispatching requests to SliceMatrix-IO

    Returns
    =======
    model : :class:`.RFRegressor`
        Trained Random Forest Regressor model

    Examples
    ========

    Train a Random Forest Regressor model

    >>> sm = SliceMatrix(api_key)
    >>> rfr = sm.RFRegressor(dataset = dataset, n_trees = 50)

    Make a prediction

    >>> rfr.predict([...])

    """
    def __init__(self, X = None, Y = None, name = None, pipeline = None, n_trees = 8, client = None):
        self.client  = client
        self.type     = "raw_rfr"
        if X is not None and Y is not None:
            self.__full_init__(X, Y, name, pipeline, n_trees, client)
        else:
            self.__lazy_init__(name)

    def __full_init__(self, X, Y, name = None, pipeline = None, n_trees = 8, client = None):
        self.n_trees = n_trees 
        if name == None:
            name = rando_name()
        # else:
        #    todo: add feature to instantiate RFRegressor just from name
        #    i.e. an already created model
        self.name     = name
        self.X        = X
        self.Y        = Y
        self.pipeline = pipeline
        if self.pipeline == None:
            pipeline_name = rando_name()
            self.pipeline = RFRegressorPipeline(pipeline_name, n_trees, client)
        self.response = self.pipeline.run(X = self.X, Y = self.Y, model = self.name)
        try:
            # model will be key if success
            model = self.response['model']
            self.name = model.split("/")[-1]
        except:
            # something went wrong creating the model
            raise StandardError(self.response)

    # lazy loading for already persisted models
    def __lazy_init__(self, model_name):
        self.name     = model_name
        
    def predict(self, point):
        """ 
        Make a prediction using the given input features.

        Also used for out of sample manifold learning.

        I.e.

        1) Perform manifold learning embedding of input data (high dimension, H) to low dimension (D, D < H), however
            - Many manifold learnin algorithms don't have straightforward out of sample generalizatons...
        2) Learn the "interpolation" function between high dim space and low dim space with a multi-output regression
            - Regress high dim (H) data points against the embedding (D) data points to learn the manifold embedding
        3) When presented with a new data point, an H dimension vector, or tensor or whatever term is fashionable,
           and "embed" it using the multi-output regression to output a D dimension vector

        Parameters
        ==========

        point : list
            List of points to use as inputs to a prediction

        Result
        ======

        prediction : list
            A list of output predictions
        """
        extra_params = {"features": point.values.tolist()}
        response = self.client.call_model(model  = self.name,
                                          type   = self.type,
                                          method = "predict",
                                          extra_params = extra_params)
        try:
           return pd.DataFrame(response['predict'])
        except:
           raise StandardError(response)

    def score(self):
        """ 
        Get the R^2 of the training dataset / predictions

        Returns
        =======

        r2 : float
            The R^2 of the training dataset
        """
        Y_hat = self.predict(self.X)
        return r_squared(Y_hat, self.Y)

# K Nearest Neighbors regressor pipeline
class KNNRegressorPipeline(BasePipeline):
    """ 
    K Nearest Neighbors Regression.

    Create a Pipeline for training :class:`KNNRegressor` models.

    Parameters
    ==========

    name : string
        The desired name of the Pipeline.

    K : integer, optional
        The desired K in the Nearest Neighbor classifier model

    kernel : string [ 'euclidean' | 'minkowski' | 'hammond' | 'etc...'], optional
        The desired kernel for defining distance in our classifier. Default is 'euclidean'

    algo : string ['auto' | 'ball' | 'kd_tree' | 'brute'], optional
        The algorithm to use in determining Nearest Neighbors. Default is 'auto'

    weights : string ['uniform' | 'weighted'], optional
        Should voting be uniform (i.e. independent of distance) or weighted by distance (i.e. closer neighbor's have higher weighted predictions)

    kernel_params : dict, optional
        Any parameters specific to the chosen kernel

    client : :class:`slicematrixIO.connect.ConnectIO`
        Low level client for dispatching requests to SliceMatrix-IO

    Returns
    =======

    response : dict
        success or failure response to Pipeline creation request

    Examples
    ========

    Create a Pipeline for training multiple :class:`.KNNRegressor` models

    >>> io = ConnectIO(api_key)
    >>> pipe = KNNRegressorPipeline(K = 5, client = io)
    >>> for dataframe in dataframes:
    >>>     current_model = pipe.run(dataset = dataframe, name = slicematrixIO.utils.rando_name())


    """
    def __init__(self, name, K = 5, kernel = "euclidean", algo = "auto", weights = "uniform", kernel_params = {}, client = None):
        params = {"k": K,
                  "kernel": kernel,
                  "algo": algo,
                  "weights": weights,
                  "kernel_params": kernel_params}
        BasePipeline.__init__(self, name, "raw_knn_regressor", client, params)

    def run(self, X, Y, model):
        """ 
        Run the Pipeline and create a new :class:`KNNRegressor` model

        Parameters
        ==========

        dataset : pandas.DataFrame
            The dataset to pass into the Pipeline which will train a :class:`KNNRegressor` model using the
            parameters defined upon Pipeline creation. Pipelines are reusable sets of instructions to train
            a machine learning model.

        Returns
        =======

        response : dict
            success or failure response to model creation request
        """
        return BasePipeline.run(self, X = X, Y = Y, model = model)

class KNNRegressor():
    """ 
    Train / Reload a :class:`.KNNRegressor` model for multi-output regression

    Parameters
    ==========

    X : pandas.DataFrame
        Input DataFrame. shape = (n_rows, input_features) where each row is a data point and the columns are numeric features

    Y : pandas.DataFrame
        Output DataFrame. shape = (n_rows, output_features) where output_features < input_features and each row is a data point
        and the columns are numeric features

    K : integer, optional
        The desired K in the Nearest Neighbor classifier model

    kernel : string [ 'euclidean' | 'minkowski' | 'hammond' | 'etc...'], optional
        The desired kernel for defining distance in our classifier. Default is 'euclidean'

    algo : string ['auto' | 'ball' | 'kd_tree' | 'brute'], optional
        The algorithm to use in determining Nearest Neighbors. Default is 'auto'

    weights : string ['uniform' | 'weighted'], optional
        Should voting be uniform (i.e. independent of distance) or weighted by distance (i.e. closer neighbor's have higher weighted predictions)

    kernel_params : dict, optional
        Any parameters specific to the chosen kernel

    name : string, optional
        The desired name of the model. If None then a random name will be generated

    pipeline : string, optional
        An extant Pipeline to use for model creation. If None then one will be created

    client : :class:`slicematrixIO.connect.ConnectIO`
        Low level client for dispatching requests to SliceMatrix-IO

    Returns
    =======
    model : :class:`.RFRegressor`
        Trained Random Forest Regressor model

    Examples
    ========

    Train a K Nearest Neighbors Regressor model

    >>> sm = SliceMatrix(api_key)
    >>> knn = sm.KNNRegressor(dataset = dataset, K = 3)

    Make a prediction

    >>> knn.predict([...])

    """
    def __init__(self, X = None, Y = None, name = None, pipeline = None, K = 5, kernel = "euclidean", algo = "auto", weights = "uniform", kernel_params = {}, client = None):
        self.client  = client
        self.type     = "raw_knn_regressor"
        if X is not None and Y is not None:
            self.__full_init__(X, Y, name, pipeline, K, kernel, algo, weights, kernel_params, client)
        else:
            self.__lazy_init__(name)

    def __full_init__(self, X, Y, name = None, pipeline = None, K = 5, kernel = "euclidean", algo = "auto", weights = "uniform", kernel_params = {}, client = None):
        self.X       = X
        self.Y       = Y
        if name == None:
            name = rando_name()
        self.name     = name
        self.pipeline = pipeline
        self.K        = K
        self.kernel   = kernel
        self.kernel_params = kernel_params
        self.algo     = algo
        self.weights  = weights
        if self.pipeline == None:
            pipeline_name = rando_name()
            self.pipeline = KNNRegressorPipeline(pipeline_name, K, kernel, algo, weights, kernel_params, client)
        self.response = self.pipeline.run(X = self.X, Y = self.Y, model = self.name)
        try:
            # model will be key if success
            model = self.response['model']
            self.name = model.split("/")[-1]
        except:
            # something went wrong creating the model
            raise StandardError(self.response)

    # lazy loading for already persisted models
    def __lazy_init__(self, model_name):
        self.name     = model_name

    def predict(self, point):
        """ 
        Make a prediction using the given input features.

        Also used for out of sample manifold learning.

        I.e.

        1) Perform manifold learning embedding of input data (high dimension, H) to low dimension (D, D < H), however
            - Many manifold learnin algorithms don't have straightforward out of sample generalizatons...
        2) Learn the "interpolation" function between high dim space and low dim space with a multi-output regression
            - Regress high dim (H) data points against the embedding (D) data points to learn the manifold embedding
        3) When presented with a new data point, an H dimension vector, or tensor or whatever term is fashionable,
           and "embed" it using the multi-output regression to output a D dimension vector

        Parameters
        ==========

        point : list
            List of points to use as inputs to a prediction

        Result
        ======

        prediction : list
            A list of output predictions
        """
        extra_params = {"features": point.values.tolist()}
        response = self.client.call_model(model  = self.name,
                                          type   = self.type,
                                          method = "predict",
                                          extra_params = extra_params)
        try:
           return pd.DataFrame(response['predict'])
        except:
           raise StandardError(response)

    def score(self):
        """ 
        Get the R^2 of the training dataset / predictions

        Returns
        =======

        r2 : float
            The R^2 of the training dataset
        """
        Y_hat = self.predict(self.X)
        return r_squared(Y_hat, self.Y)


#################################################################################################################################################################################
class KernelRidgeRegressorPipeline(BasePipeline):
    """ 
    Kernel Ridge Regression.

    Create a Pipeline for training :class:`KernelRidgeRegressor` models.

    Parameters
    ==========

    name : string
        The desired name of the Pipeline.

    alpha : float, optional
        Kernel Ridge Regressor model alpha value. Default 1.0

    kernel : string ['linear', 'rbf', 'poly']
        Kernel to use in regression. Linear is default. For nonlinear datasets, consider rbf or poly

        'linear' : linear kernel

        'rbf' : radial basis function kernel

        'poly' : polynomial kernel

    kernel_params : dict
        Kernel specific parameters

    client : :class:`slicematrixIO.connect.ConnectIO`
        Low level client for dispatching requests to SliceMatrix-IO

    Returns
    =======
    response : dict
        success or failure response to Pipeline creation request

    Examples
    ========

    Create a Pipeline for training multiple :class:`.KernelRidgeRegressor` models

    >>> io = ConnectIO(api_key)
    >>> pipe = KernelRidgeRegressorPipeline(K = 5, client = io)
    >>> for dataframe in dataframes:
    >>>     current_model = pipe.run(dataset = dataframe, name = slicematrixIO.utils.rando_name())

    """
    def __init__(self, name, kernel = "linear", alpha = 1.0, kernel_params = {}, client = None):
        params = {"kernel": kernel,
                  "alpha": alpha,
                  "kernel_params": kernel_params}
        BasePipeline.__init__(self, name, "raw_krr", client, params)

    def run(self, X, Y, model):
        """ 
        Run the Pipeline and create a new :class:`KernelRidgeRegressor` model

        Parameters
        ==========

        dataset : pandas.DataFrame
            The dataset to pass into the Pipeline which will train a :class:`KernelRidgeRegressor` model using the
            parameters defined upon Pipeline creation. Pipelines are reusable sets of instructions to train
            a machine learning model.

        Returns
        =======

        response : dict
            success or failure response to model creation request
        """
        return BasePipeline.run(self, X = X, Y = Y, model = model)

class KernelRidgeRegressor():
    """ 
    Train / Reload a :class:`.KernelRidgeRegressor` model for multi-output regression

    Parameters
    ==========

    X : pandas.DataFrame
        Input DataFrame. shape = (n_rows, input_features) where each row is a data point and the columns are numeric features

    Y : pandas.DataFrame
        Output DataFrame. shape = (n_rows, output_features) where output_features < input_features and each row is a data point
        and the columns are numeric features

    K : integer, optional
        The desired K in the Nearest Neighbor classifier model

    kernel : string [ 'euclidean' | 'minkowski' | 'hammond' | 'etc...'], optional
        The desired kernel for defining distance in our classifier. Default is 'euclidean'

    algo : string ['auto' | 'ball' | 'kd_tree' | 'brute'], optional
        The algorithm to use in determining Nearest Neighbors. Default is 'auto'

    weights : string ['uniform' | 'weighted'], optional
        Should voting be uniform (i.e. independent of distance) or weighted by distance (i.e. closer neighbor's have higher weighted predictions)

    kernel_params : dict, optional
        Any parameters specific to the chosen kernel

    name : string, optional
        The desired name of the model. If None then a random name will be generated

    pipeline : string, optional
        An extant Pipeline to use for model creation. If None then one will be created

    client : :class:`slicematrixIO.connect.ConnectIO`
        Low level client for dispatching requests to SliceMatrix-IO

    Returns
    =======
    model : :class:`.RFRegressor`
        Trained Random Forest Regressor model

    Examples
    ========

    Train a Kernel Ridge Regressor model

    >>> sm = SliceMatrix(api_key)
    >>> krr = sm.KNNRegressor(dataset = dataset, kernel = "rbf")

    Make a prediction

    >>> krr.predict([...])


    """
    def __init__(self, X = None, Y = None, name = None, pipeline = None, kernel = "linear", alpha = 1.0, kernel_params = {}, client = None):
        self.client  = client
        self.type     = "raw_krr"
        if X is not None and Y is not None:
            self.__full_init__(X, Y, name, pipeline, kernel, alpha, kernel_params, client)
        else:
            self.__lazy_init__(name)        

    def __full_init__(self, X, Y, name = None, pipeline = None, kernel = "linear", alpha = 1.0, kernel_params = {}, client = None):
        self.X       = X
        self.Y       = Y
        if name == None:
            name = rando_name()
        self.name     = name
        self.pipeline = pipeline
        self.kernel   = kernel
        self.alpha    = alpha        
        self.kernel_params = kernel_params
        if self.pipeline == None:
            pipeline_name = rando_name()
            self.pipeline = KernelRidgeRegressorPipeline(pipeline_name, kernel, alpha, kernel_params, client)
        self.response = self.pipeline.run(X = self.X, Y = self.Y, model = self.name)
        try:
            # model will be key if success
            model = self.response['model']
            self.name = model.split("/")[-1]
        except:
            # something went wrong creating the model
            raise StandardError(self.response)

    # lazy loading for already persisted models
    def __lazy_init__(self, model_name):
        self.name     = model_name

    def predict(self, point):
        """ 
        Make a prediction using the given input features. 

        Also used for out of sample manifold learning.

        I.e. 

        1) Perform manifold learning embedding of input data (high dimension, H) to low dimension (D, D < H), however
            - Many manifold learnin algorithms don't have straightforward out of sample generalizatons...
        2) Learn the "interpolation" function between high dim space and low dim space with a multi-output regression
            - Regress high dim (H) data points against the embedding (D) data points to learn the manifold embedding
        3) When presented with a new data point, an H dimension vector, or tensor or whatever term is fashionable, 
           and "embed" it using the multi-output regression to output a D dimension vector

        Parameters
        ==========

        point : list
            List of points to use as inputs to a prediction

        Result
        ======

        prediction : list
            A list of output predictions
        """
        extra_params = {"features": point.values.tolist()}
        response = self.client.call_model(model  = self.name,
                                          type   = self.type,
                                          method = "predict",
                                          extra_params = extra_params)
        try:
           return pd.DataFrame(response['predict'])
        except:
           raise StandardError(response)

    def score(self):
        """ 
        Get the R^2 of the training dataset / predictions

        Returns
        =======

        r2 : float
            The R^2 of the training dataset
        """
        Y_hat = self.predict(self.X)
        return r_squared(Y_hat, self.Y)




""" 
Classifier models are examples of supervised machine learning techniques which aim to predict the class label of a given input datapoint
"""
from core import BasePipeline
from utils import rando_name
from uuid import uuid4

# K Nearest Neighbors classifier pipeline
class KNNClassifierPipeline(BasePipeline):
    """ 
    Create a Pipeline for training :class:`.KNNClassifier` models from input datasets

    Parameters
    ==========

    name : string
        The desired name of the Pipeline.

    K : integer, optional
        The desired K in the Nearest Neighbor classifier model

    kernel : string [ 'euclidean' | 'minkowski' | 'hammond' | 'etc...'], optional
        The desired kernel for defining distance in our classifier. Default is 'euclidean'

    algo : string ['auto' | 'ball' | 'kd_tree' | 'brute'], optional
        The algorithm to use in determining Nearest Neighbors. Default is 'auto'

    weights : string ['uniform' | 'weighted'], optional
        Should voting be uniform (i.e. independent of distance) or weighted by distance (i.e. closer neighbor's have higher weighted votes)

    kernel_params : dict, optional
        Any parameters specific to the chosen kernel

    client : :class:`slicematrixIO.connect.ConnectIO`
        Low level client for dispatching requests to SliceMatrix-IO

    Returns
    =======

    response : dict
        success or failure response to Pipeline creation request

    Examples
    ========

    Create a Pipeline for training multiple :class:`.KNNClassifier` models

    >>> io = ConnectIO(api_key)
    >>> pipe = KNNClassifierPipeline(K = 7, client = io)
    >>> for dataframe in dataframes:
    >>>     current_model = pipe.run(dataset = dataframe, name = slicematrixIO.utils.rando_name())
 
    """
    def __init__(self, name, K = 5, kernel = "euclidean", algo = "auto", weights = "uniform", kernel_params = {}, client = None):
        params = {"k": K,
                  "kernel": kernel,
                  "algo": algo,
                  "weights": weights,
                  "kernel_params": kernel_params}
        BasePipeline.__init__(self, name, "raw_knn_classifier", client, params)

    def run(self, dataset, model, class_column):
        """ 
        Run the Pipeline and create a new :class:`KNNClassifier` model

        Parameters
        ==========

        dataset : pandas.DataFrame
            The dataset to pass into the Pipeline which will train a :class:`KNNClassifier` model using the
            parameters defined upon Pipeline creation. Pipelines are reusable sets of instructions to train
            a machine learning model.

        Returns
        =======

        response : dict
            success or failure response to model creation request
        """
        return BasePipeline.run(self, dataset = dataset, model = model, extra_params = {"class_column": class_column})

# K Nearest Neighbors classifier model; if pipeline == None then create a pipeline for use with this model
class KNNClassifier():
    """ 
    Train / Reload a :class:`KNNClassifier` model

    Parameters
    ==========

    dataset : pandas.DataFrame
        Input DataFrame. shape = (n_rows, n_features + 1) where each row is a data point and the columns are numeric features and a
        column with the class labels

    name : string
        The desired name of the Pipeline.

    class_column : string
        The name of the column in the input dataset which describes the class labels

    K : integer, optional
        The desired K in the Nearest Neighbor classifier model

    kernel : string [ 'euclidean' | 'minkowski' | 'hammond' | 'etc...'], optional
        The desired kernel for defining distance in our classifier. Default is 'euclidean'

    algo : string ['auto' | 'ball' | 'kd_tree' | 'brute'], optional
        The algorithm to use in determining Nearest Neighbors. Default is 'auto'

    weights : string ['uniform' | 'weighted'], optional
        Should voting be uniform (i.e. independent of distance) or weighted by distance (i.e. closer neighbor's have higher weighted votes)

    kernel_params : dict, optional
        Any parameters specific to the chosen kernel

    client : :class:`slicematrixIO.connect.ConnectIO`
        Low level client for dispatching requests to SliceMatrix-IO

    pipeline : string, optional
       An extant Pipeline to use for model creation. If None then one will be created

    Returns
    =======

    model : :class:`KNNClassifier`
        KNNClassifier model object

    Examples
    ========

    Create a KNNClassifier model for a given dataset

    >>> sm = SliceMatrix(api_key)
    >>> knn = sm.KNNClassifier(dataset = dataframe, K = 5)

    Predict the class of some new data

    >>> knn.predict([...])

    """
    def __init__(self, dataset = None, class_column = None, name = None, pipeline = None, K = 5, kernel = "euclidean", algo = "auto", weights = "uniform", kernel_params = {}, client = None):
        self.client  = client 
        self.type     = "raw_knn_classifier"
        if dataset is not None:
            self.__full_init__(dataset, class_column, name, pipeline, K, kernel, algo, weights, kernel_params, client)
        else:
            self.__lazy_init__(name)     
   
    def __full_init__(self, dataset, class_column, name = None, pipeline = None, k = 5, kernel = "euclidean", algo = "auto", weights = "uniform", kernel_params = {}, client = None):
        self.class_column = class_column
        if name == None:
            name = rando_name()
        self.name     = name
        self.dataset  = dataset
        self.pipeline = pipeline
        if self.pipeline == None:
            pipeline_name = rando_name()
            self.pipeline = KNNClassifierPipeline(pipeline_name, k, kernel, algo, weights, kernel_params, client)
        self.response = self.pipeline.run(self.dataset, self.name, self.class_column)
        try:
            # model will be key if success
            model = self.response['model']
            self.name = model.split("/")[-1]
        except:
            # something went wrong creating the model
            raise StandardError(self.response)

    # lazy loading for already persisted models
    def __lazy_init__(self, model_name):
        self.name     = model_name
        
    def predict(self, point):
        """ 
        Predict the class of new input datapoints

        Parameters
        ==========

        point : list
            A list of new datapoints. Shape = (n_points, n_features)

        Returns
        =======

        prediction : list
            A list of new predictions for each input datapoint. Shape = (n_points, 1)

        """
        extra_params = {"features": point}
        response = self.client.call_model(model  = self.name,
                                          type   = self.type,
                                          method = "predict",
                                          extra_params = extra_params)
        try:
           return response['predict']
        except:
           raise StandardError(response)

    def score(self):
        """ 
        Get the training prediction R^2 

        Returns
        =======

        r2 : float
            The R^2 of the training predictions
        """
        response = self.client.call_model(model  = self.name,
                                          type   = self.type,
                                          method = "score",
                                          extra_params = {})
        try:
           return response['score']
        except:
           raise StandardError(response)

    def training_preds(self):
        """ 
        Get the training predictions

        Returns
        =======

        prediction : list
            A list of the training predictions
        """
        response = self.client.call_model(model  = self.name,
                                          type   = self.type,
                                          method = "training_preds",
                                          extra_params = {})
        try:
           return response['training_preds']
        except:
           raise StandardError(response)

    def training_data(self):
        """ 
        Get the input data used to train the model

        Returns
        =======

        data : list
            The training data
        """
        response = self.client.call_model(model  = self.name,
                                          type   = self.type,
                                          method = "training_data",
                                          extra_params = {})
        try:
           return response['training_data']
        except:
           raise StandardError(response)

# probalistic neural network classifier pipeline
class PNNClassifierPipeline(BasePipeline):
    """ 
    Create a Pipeline for training :class:`.PNNClassifier` models from input datasets

    Parameters
    ==========

    name : string
        The desired name of the Pipeline.

    sigma : float in (0., 1.), optional
        The desired smoothing parameter for the PNN model. Default is 0.1

    client : :class:`slicematrixIO.connect.ConnectIO`
        Low level client for dispatching requests to SliceMatrix-IO

    Returns
    =======

    response : dict
        success or failure response to Pipeline creation request

    Examples
    ========

    Create a Pipeline for training multiple :class:`.PNNClassifier` models

    >>> io = ConnectIO(api_key)
    >>> pipe = PNNClassifierPipeline(sigma = 0.05, client = io)
    >>> for dataframe in dataframes:
    >>>     current_model = pipe.run(dataset = dataframe, name = slicematrixIO.utils.rando_name())

    """

    def __init__(self, name, sigma = 0.1, client = None):
        params = {"sigma": sigma}
        BasePipeline.__init__(self, name, "raw_pnn", client, params)

    def run(self, dataset, model, class_column):
        """ 
        Run the Pipeline and create a new :class:`PNNClassifier` model

        Parameters
        ==========

        dataset : pandas.DataFrame
            The dataset to pass into the Pipeline which will train a :class:`PNNClassifier` model using the
            parameters defined upon Pipeline creation. Pipelines are reusable sets of instructions to train
            a machine learning model.

        Returns
        =======

        response : dict
            success or failure response to model creation request
        """
        return BasePipeline.run(self, dataset = dataset, model = model, extra_params = {"class_column": class_column})

# probalistic neural network classifier model; if pipeline == None then create a pipeline for use with this model
class PNNClassifier():
    """ 
    Train / Reload a :class:`PNNClassifier` model

    Parameters
    ==========

    dataset : pandas.DataFrame
        Input DataFrame. shape = (n_rows, n_features + 1) where each row is a data point and the columns are numeric features and a
        column with the class labels

    name : string
        The desired name of the Pipeline.

    class_column : string
        The name of the column in the input dataset which describes the class labels


    client : :class:`slicematrixIO.connect.ConnectIO`
        Low level client for dispatching requests to SliceMatrix-IO

    pipeline : string, optional
       An extant Pipeline to use for model creation. If None then one will be created

    Returns
    =======

    model : :class:`PNNClassifier`
        PNNClassifier model object

    Examples
    ========

    Create a PNNClassifier model for a given dataset

    >>> sm = SliceMatrix(api_key)
    >>> pnn = sm.PNNClassifier(dataset = dataframe, sigma = 0.12)

    Predict the class of some new data

    >>> pnn.predict([...])

    """
    def __init__(self, dataset, class_column, name = None, pipeline = None, sigma = 0.1, client = None):
        self.client  = client
        self.class_column = class_column
        self.type     = "raw_pnn"
        if dataset is not None:
            self.__full_init__(dataset, name, pipeline, sigma, client)
        else:
            self.__lazy_init__(name)

    def __full_init__(self, dataset, name = None, pipeline = None, sigma = 0.1, client = None):
        if name == None:
            name = rando_name()
        self.name     = name
        self.dataset  = dataset
        self.pipeline = pipeline
        if self.pipeline == None:
            pipeline_name = rando_name()
            self.pipeline = PNNClassifierPipeline(pipeline_name, sigma, client)
        self.response = self.pipeline.run(self.dataset, self.name, self.class_column)
        try:
            # model will be key if success
            model = self.response['model']
            self.name = model.split("/")[-1]
        except:
            # something went wrong creating the model
            raise StandardError(self.response)

    # lazy loading for already persisted models
    def __lazy_init__(self, model_name):
        self.name     = model_name

    def predict(self, point):
        """ 
        Predict the class of new input datapoints

        Parameters
        ==========

        point : list
            A list of new datapoints. Shape = (n_points, n_features)

        Returns
        =======

        prediction : list
            A list of new predictions for each input datapoint. Shape = (n_points, 1)

        """
        extra_params = {"features": point}
        response = self.client.call_model(model  = self.name,
                                          type   = self.type,
                                          method = "predict",
                                          extra_params = extra_params)
        try:
           return response['predict']
        except:
           raise StandardError(response)

    def score(self):
        """ 
        Get the training prediction R^2

        Returns
        =======

        r2 : float
            The R^2 of the training predictions
        """
        response = self.client.call_model(model  = self.name,
                                          type   = self.type,
                                          method = "score",
                                          extra_params = {})
        try:
           return response['score']
        except:
           raise StandardError(response)

    def training_preds(self):
        """ 
        Get the training predictions

        Returns
        =======

        prediction : list
            A list of the training predictions
        """
        response = self.client.call_model(model  = self.name,
                                          type   = self.type,
                                          method = "training_preds",
                                          extra_params = {})
        try:
           return response['training_preds']
        except:
           raise StandardError(response)

    def training_data(self):
        """ 
        Get the input data used to train the model

        Returns
        =======

        data : list
            The training data
        """
        response = self.client.call_model(model  = self.name,
                                          type   = self.type,
                                          method = "training_data",
                                          extra_params = {})
        try:
           return response['training_data']
        except:
           raise StandardError(response)





""" 
Module containing all Jupyter Notebook related classes / functions

All of this is meant to be run inside a Jupyter Notebook. The resulting graphs can be shared as notebook or html. 
"""
from IPython.core.display import HTML
from IPython.display import Javascript
import json
from uuid import uuid4
from IPython.display import DisplayObject
from IPython.display import clear_output

class GraphEngine():
	""" 
	Class for setting up and drawing graphs / visualizations of SliceMatrix-I0 models directly in the Jupyter Notebook

	Parameters
	==========

        sm : :class:`slicematrixIO.client.SliceMatrix`
                An extant client

        Examples
        ========

        Create the GraphEngine

        >>> sm  = SliceMatrix(api_key)
        >>> viz = GraphEngine(sm)

        Initialize the notebook data

        >>> viz.init_data()

        Initialize the graph stylesheet

        >>> viz.init_style()

        Then visualize a model

        >>> iso = sm.Isomap(dataset=prices)
        >>> viz.drawNetworkGraph(iso, width = 1000, height = 600, color_map = "Heat")

        You can then save and export the notebook for sharing your graph. HTML exports will render directly in the browser.

        For another example check out https://slicematrix.github.io/manifold_learning_js.html
        """

	def __init__(self, sm):
		self.sm = sm
		self.graph_count = 0
		print "initializing window.graph_data"

	def drawNetworkGraph(self, network_model, color_map = "RdBuGn", graph_style = "light", graph_layout = "force", width = 1000, height = 600, charge = -100, color_axis = "closeness_centrality", label_color = "#000", label_shadow_color = "#fffff0", min_node_size = 5):
		""" 
		Embed a D3 network graph into a Jupyter Notebook to visualize a SiceMatrix-IO network graph model. Graphs embedded in notebooks can be shared. 

		Parameters
		==========

		network_model : graph-object
		    The netwok graph model. Can be any model with the methods:

		    - .nodes()
		    - .edges()
		    - .rankNodes()

		    Current list of graphable objects includes: 

		    - :class:`slicematrixIO.graphs.MinimumSpanningTree`
		    - :class:`slicematrixIO.graphs.CorrelationFilteredGraph`
		    - :class:`slicematrixIO.graphs.NeighborNetworkGraph`
		    - :class:`slicematrixIO.manifolds.Isomap`
		    - :class:`slicematrixIO.matrix_models.MatrixMinimumSpanningTree`

		color_map : string 
		    The desired color map for the node colors. Nodes are colored relative to their color_axis (the graph node statistic) selection. 

		    Mappings go from min value to median value to max value 

                    - 'RdBuGn' : Red to Blue to Green 
		    - 'RdGrGn' : Red to Gray to Green
		    - 'PuBuXr' : A purple to blue x-ray effect where nodes near the median appear to disappear on a dark background
		    - 'Viridis' : The Viridis color map, good for dark or light backgrounds 
		    - 'Heat' : A Red/Orange colormap with darker hues at the extreme
		    - 'Winter' : A Blue/Green colormap

		graph_style : string ['light' | 'dark']
                    The overall styling of the graph. Light background vs dark background...

		graph_layout : string ['force' | 'embedding']
		    The layout algorithm for the network graph. 

		    - 'force' : network layout (node positioning) will be determined by a force directed simulation
		    - 'embedding' : node positioning will be static and determined by the positions returned in model.embedding().
		      For models without a .embedding() function, enabling this option may cause the graph to fail to display properly

		width : integer, greater than 0
		    The desired width of the network graph

		height : integer, greater than 0
 		    The desired height of the network graph

		charge : integer, less than 0
		    For graph_layout == 'force', ignored otherwise. The charge associated with each node for use in the force directed simulation layout.
		    The more negative charge, the more the nodes tend to repel one another

		color_axis : string
		    The name of the graph statistic to use for coloring the graph nodes. Should be valid statistic name for call to model.rankNodes()

		label_color : string
		    The color of the node labels. Defaults to black. Accepts valid html color (e.g. #fff or rgba(255,255,255,0.8))

		label_shadow_color : string 
		    The color of the node label shadow. Defaults to "#fffff0"

		min_node_size : integer, greater than 0
		    The minimum size to make the graph nodes. Defaults to 5

		Returns
		=======

		html : IPython.display.javascript
		    A html + javascript network graph chart embedded in the Jupyter Notebook

		"""
		print self.graph_count
		clear_output()
		nodes = network_model.nodes()
		links = network_model.edges()
		stats = network_model.rankNodes(statistic=color_axis).values.tolist()
		try:
			embedding = network_model.embedding().values.tolist()
			got_embed = True
		except:
			embedding = []
			got_embed = False
		
		config = """
			require.config({
				paths: {
					d3: '//cdnjs.cloudflare.com/ajax/libs/d3/4.7.1/d3.min'
				}
			});
		"""
		
		init_js  = "element.data('graph_id', {});".format(self.graph_count);
		
		width_js  = "element.data('width', {});".format(width);
		
		height_js  = "element.data('height', {});".format(height);
		
		javascript = "window.graph_data.push(JSON.parse('{}'));".format(json.dumps({"nodes":nodes, 
																			  "links":links, 
																			  "embedding": embedding,
																			  "stats": stats,
																			  "got_embed": got_embed,
																			  "width": width,
																			  "height": height,
																			  "color_map": color_map,
																			  "graph_style": graph_style,
																			  "graph_layout": graph_layout,
																			  "charge": charge,
																			  "color_axis": color_axis,
																			  "label_color": label_color,
																			  "label_shadow_color": label_shadow_color,
																			  "min_node_size": min_node_size,
																			  "graph_id": self.graph_count}))
		#print javascript
		
		
		
		js_code = """
		window.current_graph_id += 1;
		window.graph_data[window.current_graph_id]['element'] = element;
		var graph_id = window.current_graph_id;
		console.log(graph_id);
		var width  = element.data("width");
		var height = element.data("height");
		console.log(width);
		console.log(height);
		element.empty();
		element.append('<div id = "top_bar"><b><a href = "http://www.slicematrix.com" target = "_blank" style="color:#fffff0">SliceMatrix-IO</a></b> Network View</div>');
		element.append('<div class = "loadingDialog" id = "spinner_' + graph_id +'"><div class="loader"></div><br><div id = "loading_status">loading graph data...</div></div> ');
		element.append('<div class = "content_cell" id = "content_' + graph_id +'"><canvas id = "mainEvent_' + graph_id +'" width="' + width + '" height="' + height + '"></canvas></div>');
		(function(graph_id, element) {
			console.log("inside closure!");
			require(['d3'], function(d3){
				window.d3 = d3;
				//var graph_id = element.data('graph_id');//window.current_graph_id;
				//var element  = $(element);
				
					//var element = window.graph_data[graph_id]['element'];
					console.log("require");
					console.log("graph_id = " + graph_id);
					var width = window.graph_data[graph_id]['width'];
					var height = window.graph_data[graph_id]['height'];
					console.log(window.current_graph_id);


					var label_color = window.graph_data[graph_id]['label_color'];
					
					var min_node_size = window.graph_data[graph_id]['min_node_size'];
					
					var charge = window.graph_data[graph_id]['charge'];

					var colorAxis = window.graph_data[graph_id]['color_axis'];
					if(colorAxis == undefined){
						colorAxis = "closeness_centrality";
					}

					var colorMap = window.graph_data[graph_id]['color_map'];
					if(colorMap == undefined){
						colorMap = "RdBuGn";
					}

					var graphStyle = window.graph_data[graph_id]['graph_style'];
					// light || dark || white || ether
					if(graphStyle == undefined){
						graphStyle = "light";
					}

					var suppressLabels;
					if(suppressLabels == undefined){
						suppressLabels = true;
					}

					var graphTitle;
					if(graphTitle == undefined){
						graphTitle = "SliceMatrix-IO Network Viewer";
					}

					var graphLayout = window.graph_data[graph_id]['graph_layout'];
					if(graphLayout == undefined && window.graph_data[graph_id]['got_embed'] == true){
						graphLayout = "embedding";
					}

					// update background style
					if(graphStyle == "dark"){
						d3.select("#content_" + graph_id).style("background", "rgb(20,20,20)");


						$('#top_bar').css({
							"color": "rgba(255, 255, 255, 0.88)",
							"text-shadow": "5px 5px 2px rgba(120,120,120,0.2)",
							"background-color": "rgba(220,220,220, 0.4)",
							"box-shadow": "-5px 5px 5px rgba(220,220,220,0.0)"
						});
						d3.select("#loading_status").style("color", "rgb(255, 255, 255)");
					}else if(graphStyle == "white"){
						$('#content' + graph_id).css({
							"background-image": "none",
							"background-color": "#FFF",
							"background": "rgb(256,256,256)"
						});

						d3.select("#loadingDialog").style("background", "rgba(120,120,120,0.7)");
					}else if(graphStyle == "ether"){
						$('#content' + graph_id).css({
							"background-image": "none",
							"background-color": "#000",
							"background": "-webkit-radial-gradient(circle, black 15%, rgb(5,5,10), rgb(10,10,30), rgb(15,50,50))"
						});

						$('#top_bar').css({
							"color": "rgba(255, 255, 255, 0.88)",
							"text-shadow": "5px 5px 2px rgba(120,120,120,0.2)",
							"background-color": "rgba(220,220,220, 0.4)",
							"box-shadow": "-5px 5px 5px rgba(220,220,220,0.0)"
						});

						d3.select("#loading_status").style("color", "rgb(255, 255, 255)");
					}

					//
					if(graphLayout == undefined){
						graphLayout = false;
					}

					var colormap = function(maxval, midval, minval, maxcol, midcol, mincol){
					return d3.scaleLinear()
							 .domain([minval, midval, maxval])
							 .range([mincol, midcol, maxcol]);

					}

					var colormaps = {'RdBuGn':   ['rgb(20,220,120)','rgb(20,120,220)','rgb(220,20,20)'],
								 'RdGrGn':   ['rgb(0,254,122)','rgb(185,185,185)','rgb(225,26,29)'],//2
								 'PuBuXr':   ['rgb(20,120,255)','rgb(25,25,45)','rgb(255,20,120)'],//3
								 'RdBuGnXr': ['rgb(20,220,120)','rgb(20,60,120)','rgb(220,20,20)'],//4
								 'Viridis':  ['rgb(240,249,33)','rgb(30,153,138)','rgb(19,6,137)'],//5
								 'Heat':     ['rgb(110,0,0)','rgb(255,17,0)','rgb(255,255,87)'],//9
								 'Cool':     ['rgb(6,248,255)','rgb(139,116,255)','rgb(255,0,255)'],//11
								 'Greens':   ['rgb(169,196,222)','rgb(121,198,122)','rgb(48,113,60)'],//8
								 'Blues':    ['rgb(247,252,255)','rgb(95,166,209)','rgb(19,95,167)'],//7
								 'GnBu':     ['rgb(247,252,240)','rgb(124,249,125)','rgb(8,73,138)'],//6
								 'Winter':   ['rgb(0,11,249)','rgb(0,173,169)','rgb(0,255,128)']}//10

					function getMaxOfArray(numArray) {
					  return Math.max.apply(null, numArray);
					}

					function getMinOfArray(numArray) {
					  return Math.min.apply(null, numArray);
					}

					var graph = {nodes:[], links:[]};

					var curr_colormap;

					d3.select("#loading_status").html("loaded nodes...");
					var data = window.graph_data[graph_id];
					data['nodes'].forEach(function(d){
						graph['nodes'].push({id:d});
					});

					d3.select("#loading_status").html("loaded edges...");
					data['links'].forEach(function(d){
						graph['links'].push({source:d[0], target:d[1]});
					});

					//var height = 600;
					//var width  = 1200;
					var searchRadius = 3;

					//d3.select("canvas").attr("height", height).attr("width", width);

					var transform = d3.zoomIdentity;
					
					console.log($("#mainEvent_" + graph_id));

					try{
						//var canvas = document.querySelector("canvas"),
						var canvas = document.querySelector("#mainEvent_" + graph_id),
							context = canvas.getContext("2d"),
							width = canvas.width,
							height = canvas.height;
							
						$("#mainEvent_" + graph_id).width(width);
						$("#mainEvent_" + graph_id).height(height);

						var simulation = d3.forceSimulation()
							.force("link", d3.forceLink().id(function(d) { return d.id; }))
							.force("charge", d3.forceManyBody().strength(charge))
							.force("center", d3.forceCenter(width / 2, height / 2))
							.force("X", d3.forceX().x(0.2))
							.force("Y", d3.forceY().y(0.2));

						var zoom = d3.zoom();

						var k = 1.0;

						function initialize(){
						  //zoom.scaleTo(d3.select("canvas").transition().duration(3000), 0.05);
						}

						//initialize();

						var ggraph;

						var selected_node;

						var transform;

						var ticks = 0;


						var mixcolors = function(color1, color2){
							var c1 = color1.replace("rgb(", "").replace(")", "").split(",");
							var c2 = color2.replace("rgb(", "").replace(")", "").split(",");
							var r1 = Math.round((parseInt(c1[0]) + parseInt(c2[0])) / 2);
							var g1 = Math.round((parseInt(c1[1]) + parseInt(c2[1])) / 2);
							var b1 = Math.round((parseInt(c1[2]) + parseInt(c2[2])) / 2);
							if (isNaN(r1) || isNaN(g1) || isNaN(b1)){
							  return "rgba(20,120,220,0.6)";
							}else{
							  return "rgba(" + r1.toString() + "," + g1.toString() + "," + b1.toString() + ",0.6)";
							}
						}

						var drawGraph = function(){
							console.log(graph);

							function dragsubject() {
							  var i,
								  x = transform.invertX(d3.event.x),
								  y = transform.invertY(d3.event.y),
								  dx,
								  dy;
							}


							function dragstarted() {
							  if (!d3.event.active) simulation.alphaTarget(0.3).restart();
							  d3.event.subject.fx = d3.event.subject.x;
							  d3.event.subject.fy = d3.event.subject.y;
							}

							function dragged() {
							  //d3.event.subject[0] = transform.invertX(d3.event.x);
							  //d3.event.subject[1] = transform.invertY(d3.event.y);
							  d3.event.subject.fx = d3.event.x;
							  d3.event.subject.fy = d3.event.y;
							}

							function dragended() {
							  if (!d3.event.active) simulation.alphaTarget(0);
							  d3.event.subject.fx = null;
							  d3.event.subject.fy = null;
							}

							function drawLink(d) {
							  context.moveTo(d.source.x, d.source.y);
							  context.lineTo(d.target.x, d.target.y);
							}

							function drawNode(d) {
							  context.moveTo(d.x + get_current_node_size(), d.y);
							  context.arc(d.x, d.y, get_current_node_size(), 0, 2 * Math.PI);
							  //context.fill();
							}

							function drawLabel(d, id) {
							  //console.log(d);
							  context.shadowBlur = 10;
							  context.shadowColor = "rgba(120,120,120,0.3)";
							  context.fillText(d.id, d.x,d.y);
							  //context.fill();
							}

							function rgb2a(rgb_string, alpha){
								//console.log(rgb_string);
							  rgb_string = rgb_string.replace(")", "," + alpha + ")").replace("rgb", "rgba");
							  //rgb_string = "rgba(0,0,0,0.3)";
							  return rgb_string 
							}

							function get_current_alpha(){
							  return 1.0 - Math.min(transform.k, 0.5);
							}

							function get_current_font_size(){
							  return 10 / Math.pow(transform.k, 0.5);
							}

							function get_current_highlighted_font_size(){
							  return Math.min(720 * transform.k, 20);
							}

							function get_current_link_width(){
							  return Math.max(1.5, 0.5/ transform.k);
							}

							function get_current_node_size(){
							  return Math.min(60 * transform.k, min_node_size);
							}

							function get_current_deg_thresh(){
							  if(k >= 0.4){
								return 1;
							  }else{
								return 2;
							  }
							}


							ggraph = graph;
							  setTimeout(initialize, 0);

							  if(graphLayout == "force"){

								  simulation
									  .nodes(graph.nodes)
									  .on("tick", ticked);

								  simulation.force("link")
									  .links(graph.links);

							  }else{
									  simulation
									  .nodes(graph.nodes);
									  //.on("tick", ticked);

								  simulation.force("link")
									  .links(graph.links);
									ticked();
							  }

							  //d3.select("mainEvent_" + graph_id)
							  
							  d3.select(canvas)
								  .on("mousemove", mousemoved)
								  .call(d3.drag()
									  .container(canvas)
									  .subject(dragsubject)
									  .on("start", dragstarted)
									  .on("drag", dragged)
									  .on("end", dragended))
								  .call(zoom.on("zoom", zoomed));

							  function zoomed() {
								//console.log("zooming!");
								transform = d3.event.transform;
								//console.log(simulation.alpha());
								if(graphLayout == "force"){
									if(simulation.alpha() < 0.01){//>
									  ticked();
									}
								}else{
									ticked();
								}
							  }

							  function mousemoved() {
								//console.log(graph.nodes);
								var m = d3.mouse(this);
								var x = (m[0] - transform.x) / transform.k;
								var y = (m[1] - transform.y) / transform.k;
								var moused_node;
								for (var i = graph.nodes.length - 1; i >= 0; --i) {
								  var point = graph.nodes[i];
								  var dx = x - point.x;
								  var dy = y - point.y;
								  if (dx * dx + dy * dy < searchRadius * searchRadius) {//>
									//point.x = transform.applyX(point.x);
									//point.y = transform.applyY(point.y);
									moused_node = point;
								  }
								}
								selected_node = moused_node;
								if(graphLayout == "force"){
									if(simulation.alpha() < 0.01){//>
									  ticked();
									}
								}
							  }

							  function restartIt(){
								simulation.alpha(1.0)
								simulation.restart();
							  }

							  function ticked() {
								k = transform.k;
								if( graphLayout != "force"){
									simulation.stop();
								}else{
									if(ticks == 0){
										setTimeout(restartIt,1);
									}
								}
									context.clearRect(0, 0, width, height);
									context.save();
									//context.translate(width / 2, height / 2);

									context.globalAlpha = 1.0;

									context.translate(transform.x, transform.y);
									context.scale(transform.k, transform.k);

									graph.links.forEach(function(link) {
									  context.beginPath();
									  drawLink(link);
									  //context.strokeStyle = rgb2a("rgb(20,20,20)", get_current_alpha());
									  //console.log(link.source.color);
									  var color = mixcolors(link.source.color, link.target.color);
									  context.strokeStyle = color;
									  context.lineWidth = get_current_link_width();
									  context.stroke();
									});

									graph.nodes.forEach(function(user) {
									//console.log(user);
									  context.beginPath();
									  drawNode(user);
									  context.fillStyle = rgb2a(user.color, get_current_alpha());
									  context.fill();
									  if(selected_node != undefined){
										if(user.id == selected_node.id){
										  context.strokeStyle = "#000";
										  context.lineWidth = 1;
										  context.stroke();
										}
									  }
									  //drawLabel(user, user.id);

									});

								   graph.nodes.forEach(function(user) {
									  //if(user.deg > get_current_deg_thresh()){
										context.beginPath();
										drawLabel(user, user.id);
										  context.font="bold " + get_current_font_size() + "px Arial";
										  context.fillStyle = label_color;//user.color;
										  context.fill();
									  //}
									});

									ticks++;
									context.restore();

							  }


						}

						var max_rank = 0;
						var min_rank = 0;

						window.graph_data[graph_id]['stats'].forEach(function(row, j){
							//console.log(row);
							var val1 = row[0];
							if(val1 > max_rank){
								max_rank = val1;
							}
							if(val1 < min_rank){
								min_rank = val1;
							}
							graph.nodes[j][colorAxis] = val1
						});

						d3.select("#loading_status").html("loaded colorAxis data...");

						var mid_rank = (max_rank - min_rank) / 2.0;

						//console.log([min_rank, mid_rank, max_rank, colormaps[colorMap][0], colormaps[colorMap][1], colormaps[colorMap][2]]);

						curr_colormap = colormap(min_rank, mid_rank, max_rank, colormaps[colorMap][0], colormaps[colorMap][1], colormaps[colorMap][2]);

						graph.nodes.forEach(function(node, k){
							graph.nodes[k].color = curr_colormap(graph.nodes[k][colorAxis]);
						});

						if(graphLayout == "force"){
							d3.select("#spinner_" + graph_id).style("visibility", "hidden");
							drawGraph();
							//window.current_graph_id += 1
						}else{
							// bind the node embedding data to the graph!
							var max_xy = 0.0;
							var min_xy = 0.0;
							var cmax_xy, cmin_xy;
							// normalize embedding
							//console.log("ooooooooooooooooooooooooo");
							window.graph_data[graph_id]['embedding'].forEach(function(row, j){
								cmax_xy = getMaxOfArray(row);
								cmin_xy = getMinOfArray(row);
								if(cmax_xy > max_xy){
									max_xy = cmax_xy;
								}
								if(cmin_xy < min_xy){
									min_xy = cmin_xy;
								}
							});
							//console.log("aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa")
							var range_xy = max_xy - min_xy;

							window.graph_data[graph_id]['embedding'].forEach(function(row, j){
								//console.log(row);
								graph.nodes[j]['x'] = ( ( ( Number(row[0]) ) / range_xy) * (width) )  + (width / 2);
								graph.nodes[j]['y'] = ( ( ( Number(row[1]) ) / range_xy) * (height) ) + (height / 2);
								graph.nodes[j]['fixed'] = true;
							});

							d3.select("#spinner_" + graph_id).style("visibility", "hidden");
							drawGraph();
							//window.current_graph_id += 1
						}
							
					}catch(err){
						console.log(err);
						console.log("element missing, moving on...");
					}
			});
		})(graph_id, element);
		
		"""
		self.graph_count += 1
		return Javascript(init_js + width_js + height_js + config + javascript + js_code)
		
	
	def init_data(self):
		""" 
		Initialize the window's graph data

		Returns
		=======

                js : IPython.display.javascript
                    A javascript code block
		
		"""
		# check if window.graph_data already exists
		# if so then return True
		# else create window.graph_data = {} then return false
		js_code = """
			window.graph_data = [];
			window.current_graph_id = -1;
			window.current_graph_id2 = 0; 
			
			/*
			require(['d3'], function(d3){
				window.d3 = d3;
			});
			*/
		"""
		return Javascript(js_code)
		

	def init_style(self):
                """ 
                Initialize the notebook's graph style

                Returns
                =======

                js : IPython.display.javascript
                    A javascript code block

                """

		return HTML("""
		<style>
		.content_cell{
			background-color: transparent;
			background-image:       linear-gradient(0deg, transparent 24%, rgba(0, 0, 0, .05) 25%, rgba(0, 0, 0, .05) 26%, transparent 27%, transparent 74%, rgba(0, 0, 0, .05) 75%, rgba(0, 0, 0, .05) 76%, transparent 77%, transparent), linear-gradient(90deg, transparent 24%, rgba(0, 0, 0, .05) 25%, rgba(0, 0, 0, .05) 26%, transparent 27%, transparent 74%, rgba(0, 0, 0, .05) 75%, rgba(0, 0, 0, .05) 76%, transparent 77%, transparent);
		  height:100%;
		  background-size:60px 60px;
		  z-index: -1;
		}
		#header{
		  //position: absolute;
		  //top: 0;
		  //left: 0;
		  margin: 0px;
		  //height:100px;
		  width: 100%;
		  background: #fff;
		  padding:8px;
		}


		#submit_button{
		  width:50px;
		  height:50px;
		  font-size:32px;
		}

		.ui-autocomplete {
			max-height: 400px;
			overflow-y: auto;
			/* prevent horizontal scrollbar */
			overflow-x: hidden;
		  }


		#top_bar{
			position: absolute;
			//padding: 25px;
			right: 0;
			width: auto;//200px;
			height: 20px;
			z-index: 100;
			padding-left: 15px;
			padding-right:10px;
			color: rgba(255, 255, 255, 0.88);//rgba(20, 120, 220, 0.6);

			font-size: 1.0em;
			text-shadow: 5px 5px 2px rgba(120,120,120,0.2);
			background-color: rgba(20,20,20, 0.4);
			box-shadow: 5px 5px 5px rgba(0,0,0,0.2);
			border-radius: 0px 0px 0px 25px;
		}

		.loadingDialog {
		  position: absolute;
		  padding: 10px;
		  width: 350px;
		  height: 200px;
		  z-index: 15;
		  top: 50%;
		  left: 50%;
		  margin: -100px 0 0 -150px;
		  background: rgba(40,40,40,0.7);
		  border-radius: 25px 25px 25px 25px;
		  box-shadow: 5px 5px 5px rgba(0,0,0,0.2);
		}

		.loader {
		width: 50%;
			margin: 0 auto; 
			border: 16px solid #f3f3f3; /* Light grey */
			border-top: 16px solid rgb(20,120,220);
			border-right: 16px solid rgb(20,220,120);
			border-bottom: 16px solid rgb(220,20,20);
			border-radius: 50%;
			width: 120px;
			height: 120px;
			animation: spin 2s linear infinite;
		}

		#loading_status {
		color: "#fff";
		width: 50%;
			margin: 0 auto; 
		}

		@keyframes spin {
			0% { transform: rotate(0deg); }
			100% { transform: rotate(360deg); }
		}

		</style>
		""")



""" 
Useful utility functions
"""
from uuid import uuid4
import numpy as np

def rando_name(type = "short"):
    """ 
    Generate a random name string

    Longer name decreases chance of overwrite collision

    Parameters
    ==========

    type : string ["short" | "long"]
        Whether to create a long or short name

    Returns
    =======

    name : string
        Random name
    """
    name = str(uuid4())
    if type == "short":
        return name.split("-")[-1]
    else:
        return name.replace("-", "")


def r_squared(Y_hat, Y):
    """ 
    Get the coefficient of determination, or r-squared, for a given prediction versus its ground truths

    Parameters
    ==========

    Y_hat : pandas.DataFrame
        The predicted values DataFrame

    Y : pandas.DataFrame
        The actual values DataFrame

    Returns
    =======

    r_2 : float
        The r-squared value
    """
    res = np.sum(np.square(np.subtract(Y.values, Y_hat.values))) #np.sum(np.square(Y.values - Y_hat.values))
    tot = np.sum(np.square(np.subtract(Y.values, np.mean(Y.values))))
    #print res, tot
    return 1. - ( res / tot )



"""High Level Python Client for the SliceMatrix-IO Machine Learning PaaS"""
from connect import ConnectIO, Uploader

from bayesian_filters import KalmanOLS                                                     # lazy loading enabled
from distributions import KernelDensityEstimator, BasicA2D, IsolationForest                # lazy loading enabled                
from classifiers import KNNClassifier, PNNClassifier                                       # lazy loading enabled
from graphs import MinimumSpanningTree, CorrelationFilteredGraph, NeighborNetworkGraph     # lazy loading enabled 
from matrices import DistanceMatrix                                                        # lazy loading enabled
from matrix_models import MatrixMinimumSpanningTree, MatrixKernelPCA, MatrixAgglomerator   # lazy loading enabled
from manifolds import KernelPCA, LocalLinearEmbedder, LaplacianEigenmapper, Isomap         # lazy loading enabled
from regressors import KNNRegressor, RFRegressor, KernelRidgeRegressor                     # lazy loading enabled

class SliceMatrix():
    """ 
    Main business object for slicematrixIO-python

    Builds upon low level api (ConnectIO) to create high level objects for each model type.

    The models are meant to be created by the client, as opposed to instantiated directly.

    Parameters
    ==========
    api_key: string
        A Valid SliceMatrix-IO API Key

    region : string ['us-east-1', 'us-west-1', 'eu-central-1', 'ap-southeast-1']
        Data center of choice. API Key must be valid for that specific data center.
        Latency will be lowest if client is closest to data center.

        'us-east-1': US East Coast Data Center

        'us-west-1': US West Coast Data Center

        'eu-central-1': Continental Europe Data Center

        'ap-southeast-1': South-East Asian Data Center

    Attributes
    ==========
    client : ConnectIO
        Low level SliceMatrix-IO Python client      

    Examples
    ========

    Create a Kernel Density Estimator model that lives in the cloud

    >>> kde = sm.KernelDensityEstimator(dataset=df) 

    Score a new data point

    >>> kde.score(10325632)

    Simulate 1000 new data points

    >>> kde.simulate(1000) 

    Manifold Learning:

    >>> iso = sm.Isomap(dataset=prices)

    Get statistics / factors related to internal graph structure of each node

    >>> iso.rankNodes("pagerank")

    Find low dimensional embedding of input data

    >>> iso.embedding() 

    See www.slicematrix.com/use-cases for more in-depth examples

    """
    def __init__(self, api_key, region = "us-east-1"):
        self.api_key = api_key
        self.client  = ConnectIO(self.api_key, region = region)

    # bayesian filters ##########################################################################################################################################################
    # need to test this
    def KalmanOLS(self, dataset = None, init_alpha = None, init_beta = None, trans_cov = None, obs_cov = None, init_cov = None, optimizations = [], name = None, pipeline = None):
        """ 
        Create slicematrixIO.bayesian_filters.KalmanOLS object with current client

        The KalmanOLS model 

        Parameters
        ==========

        dataset: pandas.DataFrame
            Input DataFrame. shape = (nrows, 2) where the first column is Y and the second is X in OLS model

        init_alpha : float, optional
            Initial value for alpha in OLS model (ignored if optimizations are enabled)

        init_beta : float, optional
            Initial value for beta in OLS model (ignored if optimizations are enabled)

        trans_cov : array-like, optional
            Transition covariance, shape = (2, 2) 

        init_cov : array-like, optional
            Initial covariance, shape = (2, 2)

        optimizations : list, optional
            List of optimizations. Can include multiple optimizations. Default includes all:

            - 'transition_covariance'
            - 'observation_covariance'
            - 'initial_state_mean'
            - 'initial_state_covariance'

        name : string, optional
            The desired name of the model. If None then a random name will be generated

        pipeline : BasePipeline, optional
            Pipeline to use. Defaults to None. If None then a pipeline will be created for use in creating the model

        Returns
        =======

        model : :class:`slicematrixIO.bayesian_filters.KalmanOLS`

        """
        return KalmanOLS(dataset = dataset,
                         init_alpha = init_alpha,
                         init_beta  = init_beta,
                         trans_cov  = trans_cov,
                         obs_cov  = obs_cov,
                         init_cov = init_cov,
                         optimizations = optimizations,
                         name     = name,
                         pipeline = pipeline,
                         client   = self.client)

    # classifiers ###############################################################################################################################################################
    def KNNClassifier(self, dataset = None, class_column = None, name = None, pipeline = None, K = 5, kernel = "euclidean", algo = "auto", weights = "uniform", kernel_params = {}):
        """ 
        Create K Nearest Neighbors Classifier model

        Parameters
        ==========

        dataset : pandas.DataFrame
            Input DataFrame. shape = (n_rows, n_features + 1) where each row is a data point and the columns are numeric features and a 
            column with the class labels

        class_column : string
            The name of the column in the input dataset which describes the class labels

        name : string, optional
            The desired name of the model. If None then a random name will be generated

        pipeline : string, optional
            An extant Pipeline to use for model creation. If None then one will be created

        K : integer, optional
            The desired K in the Nearest Neighbor classifier model

        kernel : string [ 'euclidean' | 'minkowski' | 'hammond' | 'etc...'], optional
            The desired kernel for defining distance in our classifier. Default is 'euclidean'

        algo : string ['auto' | 'ball' | 'kd_tree' | 'brute'], optional
            The algorithm to use in determining Nearest Neighbors. Default is 'auto'

        weights : string ['uniform' | 'weighted'], optional 
            Should voting be uniform (i.e. independent of distance) or weighted by distance (i.e. closer neighbor's have higher weighted votes)

        kernel_params : dict, optional
            Any parameters specific to the chosen kernel

        Returns
        =======

        model : :class:`slicematrixIO.classifiers.KNNClassifier`
        
        """
        return KNNClassifier(dataset  = dataset, 
                             class_column = class_column,
                             name     = name, 
                             pipeline = pipeline, 
                             K        = K,
                             kernel   = kernel, 
                             algo     = algo, 
                             weights  = weights, 
                             kernel_params = kernel_params, 
                             client   = self.client)

    def PNNClassifier(self, dataset = None, class_column = None, name = None, pipeline = None, sigma = 0.1):
        """ 
        Create a Probabilistic Neural Network Classifier model

        Parameters
        ==========

        dataset : pandas.DataFrame
            Input DataFrame. shape = (n_rows, n_features + 1) where each row is a data point and the columns are numeric features and a
            column with the class labels

        class_column : string
            The name of the column in the input dataset which describes the class labels

        name : string, optional
            The desired name of the model. If None then a random name will be generated

        pipeline : string, optional
            An extant Pipeline to use for model creation. If None then one will be created

        sigma : float in (0., 1.), optional
            The desired smoothing parameter for the PNN model. Default is 0.1

        Returns
        =======

        model : :class:`slicematrixIO.classifiers.PNNClassifier`
        """
        return PNNClassifier(dataset  = dataset,
                             class_column = class_column,
                             name     = name,
                             pipeline = pipeline,
                             sigma    = sigma,
                             client   = self.client)

    
    # distributions #############################################################################################################################################################
    def KernelDensityEstimator(self, dataset = None, bandwith = "scott", kernel_params = {}, name = None, pipeline = None):
        """ 
        Train a Kernel Density Estimator model 

        Parameters
        ==========        

        dataset : pandas.DataFrame
            Input DataFrame. shape = (n_rows, n_features + 1) where each row is a data point and the columns are numeric features and a
            column with the class labels

        name : string, optional
            The desired name of the model. If None then a random name will be generated

        pipeline : string, optional
            An extant Pipeline to use for model creation. If None then one will be created

        bandwidth : str ['scott' | 'silverman'], optional
            The method for bandwidth selection in the KDE model

        kernel_params : dict, optional
            Any parameters specific to the chosen kernel

        Returns
        =======
        model : :class:`slicematrixIO.distributions.KernelDensityEstimator`
 
        """
        return KernelDensityEstimator(dataset  = dataset,
                                      bandwidth = bandwith,
                                      name     = name,
                                      pipeline = pipeline,
                                      client   = self.client)

    def BasicA2D(self, dataset = None, retrain = True, name = None, pipeline = None):
        """ 
        Create a BasicA2D model (basic automatic anomaly detection)

        Parameters
        ==========

        dataset : pandas.DataFrame
            Input DataFrame. shape = (n_rows, n_features + 1) where each row is a data point and the columns are numeric features and a
            column with the class labels

        name : string, optional
            The desired name of the model. If None then a random name will be generated

        pipeline : string, optional
            An extant Pipeline to use for model creation. If None then one will be created

        retrain : boolean, optional
            Whether to automatically retrain the model upon a remote call to the update method. The BasicA2D is a window detector which
            can be retrained in an online fashion where new data is used to update the model's understanding of the world and influence
            future anomaly scoring. 

        Returns
        =======
        model : :class:`slicematrixIO.distributions.BasicA2D`
        
        """
        return BasicA2D(dataset  = dataset,
                        retrain  = retrain,
                        name     = name,
                        pipeline = pipeline,
                        client   = self.client)

    def IsolationForest(self, dataset = None, rate = 0.1, n_trees = 100, name = None, pipeline = None):
        """ 
        Create an Isolation Forest model for automatic anomaly detection

        Parameters
        ==========

        dataset : pandas.DataFrame
            Input DataFrame. shape = (n_rows, n_features + 1) where each row is a data point and the columns are numeric features and a
            column with the class labels

        name : string, optional
            The desired name of the model. If None then a random name will be generated

        pipeline : string, optional
            An extant Pipeline to use for model creation. If None then one will be created

        rate : float in (0., 0.5), optional
            The desired rate of anomaly detection in training data. Default is 0.1 i.e. 10%

        n_trees : integer greater than 1, optional
            The number of trees to use in construction of the Isolation Forest model. Default is 100 trees

         Returns
        =======
        model : :class:`slicematrixIO.distributions.IsolationForest`

        """
        return IsolationForest(dataset  = dataset,
                               rate  = rate,
                               n_trees = n_trees,
                               name     = name,
                               pipeline = pipeline,
                               client   = self.client)

    # regressors ################################################################################################################################################################
    def KNNRegressor(self, X = None, Y = None, K = 5, kernel = "euclidean", algo = "auto", weights = "uniform", kernel_params = {}, name = None, pipeline = None):
        """ 
        Create a K Nearest Neighbors Regressor model

        Multi-output regression finds function from input space Y to lower dimension output space

        Parameters
        ==========

        X : pandas.DataFrame
            Input DataFrame. shape = (n_rows, input_features) where each row is a data point and the columns are numeric features

        Y : pandas.DataFrame
            Output DataFrame. shape = (n_rows, output_features) where output_features < input_features and each row is a data point
            and the columns are numeric features

        name : string, optional
            The desired name of the model. If None then a random name will be generated

        pipeline : string, optional
            An extant Pipeline to use for model creation. If None then one will be created

        K : integer, optional
            The desired K in the Nearest Neighbor classifier model

        kernel : string [ 'euclidean' | 'minkowski' | 'hammond' | 'etc...'], optional
            The desired kernel for defining distance in our classifier. Default is 'euclidean'

        algo : string ['auto' | 'ball' | 'kd_tree' | 'brute'], optional
            The algorithm to use in determining Nearest Neighbors. Default is 'auto'

        weights : string ['uniform' | 'weighted'], optional
            Should voting be uniform (i.e. independent of distance) or weighted by distance (i.e. closer neighbor's have higher weighted predictions)

        kernel_params : dict, optional
            Any parameters specific to the chosen kernel
        
      
        Returns
        =======
        model : :class:`slicematrixIO.regressors.KNNRegressor`

        """
        return KNNRegressor(X = X,
                            Y = Y,
                            name     = name,
                            pipeline = pipeline,
                            K        = K,
                            kernel   = kernel,
                            algo     = algo,
                            weights  = weights,
                            kernel_params = kernel_params,
                            client   = self.client)


    def RFRegressor(self, X = None, Y = None, n_trees = 8, name = None, pipeline = None):
        """ 
        Create a Random Forest Regressor model

        A Random Forest Regressor finds a function which maps the input space (X) to the lower dimension output space (Y) using decision trees

        Parameters
        ==========
        X : pandas.DataFrame
            Input DataFrame. shape = (n_rows, input_features) where each row is a data point and the columns are numeric features

        Y : pandas.DataFrame
            Output DataFrame. shape = (n_rows, output_features) where output_features < input_features and each row is a data point
            and the columns are numeric features

        n_trees : integer greater than 1, optional
            The number of trees to use in construction of the Random Forest model. Default is 8 trees

        name : string, optional
            The desired name of the model. If None then a random name will be generated

        pipeline : string, optional
            An extant Pipeline to use for model creation. If None then one will be created        

        Returns
        =======
        model : :class:`slicematrixIO.regressors.RFRegressor`
        """
        return RFRegressor(X = X, 
                           Y = Y, 
                           n_trees = n_trees, 
                           name = name,
                           pipeline = pipeline,
                           client = self.client)

    def KernelRidgeRegressor(self, X = None, Y = None, kernel = "linear", alpha = 1.0, kernel_params = {}, name = None, pipeline = None):
        """ 
        Create a Kernel Ridge Regressor model

        Parameters
        ==========

        X : pandas.DataFrame
            Input DataFrame. shape = (n_rows, input_features) where each row is a data point and the columns are numeric features

        Y : pandas.DataFrame
            Output DataFrame. shape = (n_rows, output_features) where output_features < input_features and each row is a data point
            and the columns are numeric features

        alpha : float, optional
            Kernel Ridge Regressor model alpha value. Default 1.0
 
        kernel : string ['linear', 'rbf', 'poly']
            Kernel to use in regression. Linear is default. For nonlinear datasets, consider rbf or poly
        
            'linear' : linear kernel 

            'rbf' : radial basis function kernel

            'poly' : polynomial kernel

        kernel_params : dict
            Kernel specific parameters

        name : string, optional
            The desired name of the model. If None then a random name will be generated

        pipeline : string, optional
            An extant Pipeline to use for model creation. If None then one will be created

        Returns
        =======
        model : :class:`slicematrixIO.regressors.KernelRidgeRegressor`

        """
        return KernelRidgeRegressor(X = X,
                                    Y = Y,
                                    kernel = kernel, 
                                    alpha  = alpha,
                                    kernel_params = kernel_params,
                                    name = name,
                                    pipeline = pipeline,
                                    client = self.client)
         
    # matrices #################################################################################################################################################################
    def DistanceMatrix(self, dataset = None, K = 5, kernel = "euclidean", kernel_params = {}, geodesic = False, name = None, pipeline = None):
        """ 
        Create a Distance Matrix model  

        Parameters
        ==========

        dataset : pandas.DataFrame
            Input DataFrame. shape = (n_rows, n_features + 1) where each row is a data point and the columns are numeric features and a
            column with the class labels

        name : string, optional
            The desired name of the model. If None then a random name will be generated

        pipeline : string, optional
            An extant Pipeline to use for model creation. If None then one will be created

        K : integer greater than 1, optional, ignored if geodesic == False
            The number of neighbors to use in building the geodesic distance matrix. Geodesic distance is constructed by computing the
            K Nearest Neighbors graph for the input dataset, then constructing all pairwise distances using the geodesic distance, i.e.
            the number of edges in a shortest path between two points on the graph. 

        kernel : string, optional
            The distance kernel / metric to use in constructing the distance matrix. Default is euclidean.

        kernel_params : dict, optional
            Any extra parameters specific to the chosen kernel

        geodesic : boolean, optional
            Whether to create the geodesic distance matrix or the brute force pairwise distance matrix. Default is False

        Returns
        =======

        model : :class:`slicematrixIO.matrices.DistanceMatrix`

        """
        return DistanceMatrix(dataset = dataset,
                              K       = K,
                              kernel  = kernel,
                              kernel_params = kernel_params,
                              geodesic = geodesic,
                              name = name,
                              pipeline = pipeline,
                              client   = self.client)
    
    # matrix models ############################################################################################################################################################
    def MatrixMinimumSpanningTree(self, matrix = None, matrix_name = None, matrix_type = None, name = None, pipeline = None):
        """ 
        Create a Minimum Spanning Tree model from a Distance Matrix model

        This is an example of a Matrix Model, which creates a machine learning model using another already trained model as
        its input. You can think of this as model chaining. 

        In this case, this function takes a previously created Distance Matrix model and uses it to construct a network graph
        model called a Minimum Spanning Tree.

        Parameters
        ==========

        matrix : object
            Matrix model object; i.e. a class from slicematrixIO.matrices

        matrix_name : string, optional
            The name of the existing matrix model. Optional if matrix is not None

        matrix_type :
            The type of the matrix model. Optional if matrix is not None. Required if using matrix_name

        name : string, optional
            The desired name of the model. If None then a random name will be generated

        pipeline : string, optional
            An extant Pipeline to use for model creation. If None then one will be created

        Returns
        =======

        model : :class:`slicematrixIO.matrix_models.MatrixMinimumSpanningTree`

        """
        return MatrixMinimumSpanningTree(matrix = matrix,
                                         matrix_name = matrix_name,
                                         matrix_type = matrix_type,
                                         name = name,
                                         pipeline = pipeline,
                                         client   = self.client)

    def MatrixKernelPCA(self, D = 2, matrix = None, matrix_name = None, matrix_type = None, name = None, pipeline = None):
        """ 
        Decompose the input matrix and embed the input into a lower dimension space

        Parameters
        ==========

        D : int, optional
            The desired embedding dimension. Defaults to 2-D

        matrix : object
            Matrix model object; i.e. a class from slicematrixIO.matrices

        matrix_name : string, optional
            The name of the existing matrix model. Optional if matrix is not None

        matrix_type :
            The type of the matrix model. Optional if matrix is not None. Required if using matrix_name

        name : string, optional
            The desired name of the model. If None then a random name will be generated

        pipeline : string, optional
            An extant Pipeline to use for model creation. If None then one will be created

        Returns
        =======

        model : :class:`slicematrixIO.matrix_models.MatrixKernelPCA`

        """
        return MatrixKernelPCA(D = D, 
                               matrix = matrix,
                               matrix_name = matrix_name,
                               matrix_type = matrix_type,
                               name = name,
                               pipeline = pipeline,
                               client   = self.client)

    def MatrixAgglomerator(self, label_dataset = None, alpha = 0.1, matrix = None, matrix_name = None, matrix_type = None, name = None, pipeline = None):
        """ 
        Create a Matrix Agglomerator model. Essential for supervised manifold learning, this model takes a previously created matrix model as input and
        applies class label information to the similarity matrix. In a nutshell, this model pulls data points of the same class closer together, increasing
        the separability of the dataset. 

        Parameters
        ==========

        label_dataset : pandas.DataFrame
            The class label information. shape = (n_rows, 1). n_rows should be same dimension as input matrix

        alpha : float in (0., 1.)
            The agglomeration factor, i.e. how much does class label effect input distances. An alpha of 0 will
            have no effect, while 1.0 will pull data points of the same class completely together. The higher
            the value of alpha, the more information will be transfered to the distance matrix from the class
            labels. Higher alphas increase the in-sample performance but also increase the chance of over-fitting

        matrix : object
            Matrix model object; i.e. a class from :mod:`slicematrixIO.matrices`

        matrix_name : string, optional
            The name of the existing matrix model. Optional if matrix is not None

        matrix_type :
            The type of the matrix model. Optional if matrix is not None. Required if using matrix_name

        name : string, optional
            The desired name of the model. If None then a random name will be generated

        pipeline : string, optional
            An extant Pipeline to use for model creation. If None then one will be created
        
        Returns
        =======

        model : :class:`slicematrixIO.matrix_models.MatrixAgglomerator`
        """
        return MatrixAgglomerator(alpha = alpha, 
                                  label_dataset = label_dataset,
                                  matrix = matrix,
                                  matrix_name = matrix_name,
                                  matrix_type = matrix_type,
                                  name = name,
                                  pipeline = pipeline,
                                  client   = self.client)

    # manifolds ################################################################################################################################################################
    def KernelPCA(self, dataset = None, D = 2, kernel = "linear", alpha = 1.0, invert = False, kernel_params = {}, name = None, pipeline = None):
        """ 
        Create a Kernel Principal Components Analysis model for non-linear dimensionality reduction. Applies the kernel trick to PCA

        Parameters
        ==========

        dataset : pandas.DataFrame
            Input DataFrame. shape = (n_rows, n_features) where each row is a data point and the columns are numeric features

        D : int, optional
            The desired embedding dimension. Defaults to 2-D

        kernel : string, optional
            The distance kernel / metric to use in constructing the distance matrix. Default is euclidean.

        alpha : float, optional
            Parameter of ridge regression which learns the inverse transform. Ignored if invert == False

        invert : boolean, optional
            Whether to learn the inverse transform (from low dimension space back to high dimension space)

        kernel_params : dict, optional
            Any extra parameters specific to the chosen kernel        

        name : string, optional
            The desired name of the model. If None then a random name will be generated

        pipeline : string, optional
            An extant Pipeline to use for model creation. If None then one will be created

        Returns
        =======

        model : :class:`slicematrixIO.manifolds.KernelPCA`
        
        """
        return KernelPCA(dataset = dataset,
                         D = D,
                         kernel  = kernel,
                         kernel_params = kernel_params,
                         alpha = alpha,
                         invert = invert,
                         name = name,
                         pipeline = pipeline,
                         client   = self.client)

    def LocalLinearEmbedder(self, dataset = None, D = 2, K = 3, method = "standard", name = None, pipeline = None):
        """ 
        Create a Local Linear Embedder model for non-linear dimensonality reduction

        Parameters
        ==========

        dataset : pandas.DataFrame
            Input DataFrame. shape = (n_rows, n_features) where each row is a data point and the columns are numeric features

        D : int, optional
            The desired embedding dimension. Defaults to 2-D

        K : integer greater than 1, optional
            The number of neighbors to use in building the embedding. Default is 3

        method : string ['standard' | 'hessian' | 'modified' | 'ltsa']
            Which LLE algorithm should we use?

            'standard' : standard LLE method

            'hessian': hessian eigenmap LLE method, requires that K > D * (1 + (D + 1) / 2

            'modified' : modified LLE method

            'ltsa': local tangent space alignment LLE method

        name : string, optional
            The desired name of the model. If None then a random name will be generated

        pipeline : string, optional
            An extant Pipeline to use for model creation. If None then one will be created

        Returns
        =======

        model : :class:`slicematrixIO.manifolds.LocalLinearEmbedder`

        """                               

        return LocalLinearEmbedder(dataset = dataset,
                                   D = D,
                                   K = K,
                                   method = method, 
                                   name = name, 
                                   pipeline = pipeline,
                                   client   = self.client)

    def LaplacianEigenmapper(self, dataset = None, D = 2, affinity = "knn", K = 5, gamma = 1.0, name = None, pipeline = None):
        """ 
        Create Laplacian Eiegenmapper model aka spectral embedder for non-linear dimensonality reduction

        Parameters
        ==========

        dataset : pandas.DataFrame
            Input DataFrame. shape = (n_rows, n_features) where each row is a data point and the columns are numeric features

        D : int, optional
            The desired embedding dimension. Defaults to 2-D

        affinity : string ["knn" | "rbf"], optional
            How should we construct the affinity matrix? 

            "knn" : use k nearest neighbors graph
            
            "rbf" : use radial basis function kernel

        K : integer greater than 1, optional
            The K to use if affinity is "knn".

        gamma : float, optional
            Kernel coefficient for affinity "rbf"

        name : string, optional
            The desired name of the model. If None then a random name will be generated

        pipeline : string, optional
            An extant Pipeline to use for model creation. If None then one will be created

        Returns
        =======

        model : :class:`slicematrixIO.manifolds.LaplacianEigenmapper`

        """
        return LaplacianEigenmapper(dataset = dataset,
                                    D = D,
                                    K = K,
                                    gamma = gamma,
                                    name = name,
                                    pipeline = pipeline,
                                    client   = self.client)

    def Isomap(self, dataset = None, D = 2, K = 3, name = None, pipeline = None):
        """ 
        Create Isomap model for non-linear dimensonality reduction

        Parameters
        ==========

        dataset : pandas.DataFrame
            Input DataFrame. shape = (n_rows, n_features) where each row is a data point and the columns are numeric features

        D : int, optional
            The desired embedding dimension. Defaults to 2-D

        K : integer greater than 1, optional, ignored if geodesic == False
            The number of neighbors to use in building the geodesic distance matrix. Geodesic distance is constructed by computing the
            K Nearest Neighbors graph for the input dataset, then constructing all pairwise distances using the geodesic distance, i.e.
            the number of edges in a shortest path between two points on the graph.

        name : string, optional
            The desired name of the model. If None then a random name will be generated

        pipeline : string, optional
            An extant Pipeline to use for model creation. If None then one will be created

        Returns
        =======

        model : :class:`slicematrixIO.manifolds.Isomap`

        """
        return Isomap(dataset = dataset,
                      D = D,
                      K = K,
                      name = name,
                      pipeline = pipeline,
                      client   = self.client)

    # graphs ##################################################################################################################################################################
    def MinimumSpanningTree(self, dataset = None, corr_method = "pearson", name = None, pipeline = None):
        """ 
        Create a Minimum Spanning Tree graph model

        MST models transform the input dataset into a distance matrix then construct a graph with the shortest possible total
        distance which visits all nodes without cycling (i.e. it creates a tree)

        In particular, this model constructs the graph using the correlation matrix. For more flexible options in creating a 
        MST graph, use slicematrixIO.matrix_models.MatrixMinimumSpanningTree in combination with a distance matrix

        Parameters
        ==========

        dataset : pandas.DataFrame
            Input DataFrame. shape = (n_rows, n_features) where each row is a data point and the columns are numeric features

        corr_method : string ["pearson" | "spearman" | "kendall" ]
            Which method should we use for computing the correlation matrix?

            "pearson" : use the Pearson correlation coefficient

            "spearman" : use Spearman's rho

            "kendall" : use Kendall's tau

        name : string, optional
            The desired name of the model. If None then a random name will be generated

        pipeline : string, optional
            An extant Pipeline to use for model creation. If None then one will be created

        Returns
        =======

        model : :class:`slicematrixIO.graphs.MinimumSpanningTree`

        """
        return MinimumSpanningTree(dataset = dataset,
                                   corr_method = corr_method, 
                                   name = name,
                                   pipeline = pipeline,
                                   client   = self.client)
    
    def CorrelationFilteredGraph(self, dataset = None, K = 3, name = None, pipeline = None):
        """ 
        Create a Correlation Filtered Graph

        CFG are similar to MST's, in that both graph's begin with a distance matrix, but whereas 
        MST's are limited to constructing a tree, CFG's draw links between a node and its closests K
        neighbors based on correlation distance. CFG's are like KNN networks, but optimized for 
        using correlation distance. 

        Parameters
        ==========

        dataset : pandas.DataFrame
            Input DataFrame. shape = (n_rows, n_features) where each row is a data point and the columns are numeric features

        K : integer greater than 1, optional
            The number of nearest neighbors to use for constructing the CFG

        name : string, optional
            The desired name of the model. If None then a random name will be generated

        pipeline : string, optional
            An extant Pipeline to use for model creation. If None then one will be created

        Returns
        =======

        model : :class:`slicematrixIO.graphs.CorrelationFilteredGraph`

        """
        return CorrelationFilteredGraph(dataset = dataset,
                                        K = K,
                                        name = name,
                                        pipeline = pipeline,
                                        client   = self.client)

    def NeighborNetworkGraph(self, dataset = None, K = 3, kernel = "euclidean", name = None, pipeline = None):
        """ 
        Create a K Nearest Neighbor Graph for the given dataset

        Parameters
        ==========

        dataset : pandas.DataFrame
            Input DataFrame. shape = (n_rows, n_features) where each row is a data point and the columns are numeric features

        K : integer greater than 1, optional
            The number of nearest neighbors to use for constructing the CFG

        kernel : string, optional
            The distance kernel / metric to use in constructing the distance matrix. Default is euclidean.

        name : string, optional
            The desired name of the model. If None then a random name will be generated

        pipeline : string, optional
            An extant Pipeline to use for model creation. If None then one will be created

        Returns
        =======

        model : :class:`slicematrixIO.graphs.NeighborNetworkGraph`

        """
        return NeighborNetworkGraph(dataset = dataset,
                                    K = K,
                                    kernel = kernel,
                                    name = name,
                                    pipeline = pipeline,
                                    client   = self.client)

